import unittest
import requests_mock
from test.utils import validate_list
from parsons.phone2action import Phone2Action
import os

adv_json = {
    "data": [
        {
            "id": 7125439,
            "prefix": "Mr.",
            "firstname": "Bob",
            "middlename": "Smith",
            "lastname": "Smit",
            "suffix": None,
            "notes": None,
            "stage": None,
            "connections": 3,
            "tags": [
                "register-to-vote-38511",
                "registered-to-vote-for-2018-38511",
            ],
            "created_at": {
                "date": "2017-05-23 23:36:04.000000",
                "timezone_type": 3,
                "timezone": "UTC"
            },
            "updated_at": {
                "date": "2018-12-17 21:55:24.000000",
                "timezone_type": 3,
                "timezone": "UTC"
            },
            "address": {
                "street1": "25255 Maine Ave",
                "street2": "",
                "city": "Los Angeles",
                "state": "CA",
                "zip5": 96055,
                "zip4": 9534,
                "county": "Tehama",
                "latitude": "50.0632635",
                "longitude": "-122.09654"
            },
            "districts": {
                "congressional": "1",
                "stateSenate": "4",
                "stateHouse": "3",
                "cityCouncil": None
            },
            "ids": [],
            "memberships": [
                {
                    "id": 15151443,
                    "campaignid": 25373,
                    "name": "20171121 Businesses for Responsible Tax Reform - Contact Congress",
                    "source": None,
                    "created_at": {
                        "date": "2017-11-21 23:28:30.000000",
                        "timezone_type": 3,
                        "timezone": "UTC"
                    }
                },
                {
                    "id": 20025582,
                    "campaignid": 32641,
                    "name": "20180524 March for America",
                    "source": None,
                    "created_at": {
                        "date": "2018-05-24 21:09:49.000000",
                        "timezone_type": 3,
                        "timezone": "UTC"
                    }
                }
            ],
            "fields": [],
            "phones": [
                {
                    "id": 10537860,
                    "address": "+19995206447",
                    "subscribed": 'false'
                }
            ],
            "emails": [
                {
                    "id": 10537871,
                    "address": "N@k.com",
                    "subscribed": 'false'
                },
                {
                    "id": 10950446,
                    "address": "email@me.com",
                    "subscribed": 'false'
                }
            ]
        }
    ],
    "pagination": {
        "count": 1,
        "per_page": 100,
        "current_page": 1,
        "next_url": "https://api.phone2action.com/2.0/advocates?page=2"}
}

camp_json = [
    {
        "id": 25373,
        "name": "20171121 Businesses for Responsible Tax Reform - Contact Congress",
        "display_name": "Businesses for Responsible Tax Reform",
        "subtitle": "Tell Congress: Stand up for responsible tax reform!",
        "public": 1,
        "topic": None,
        "type": "CAMPAIGN",
        "link": "http://p2a.co/KHcUyTK",
        "restrict_allow": None,
        "content": {
            "summary": "",
            "introduction": "Welcome",
            "call_to_action": "Contact your officials in one click!",
            "thank_you": "<p>Thanks for taking action. Please encourage others to act by sharing on social media.</p>",
            "background_image": None
        },
        "updated_at": {
            "date": "2017-11-21 23:27:11.000000",
            "timezone_type": 3,
            "timezone": "UTC"
        }
    }
]


class TestP2A(unittest.TestCase):

    def setUp(self):

        self.p2a = Phone2Action(app_id='an_id', app_key='app_key')

    def tearDown(self):

        pass

    def test_init_args(self):
        # Test initializing class with args
        # Done in the setUp

        pass

    def test_init_envs(self):
        # Test initilizing class with envs

        os.environ['PHONE2ACTION_APP_ID'] = 'id'
        os.environ['PHONE2ACTION_APP_KEY'] = 'key'

        p2a_envs = Phone2Action()
        self.assertEqual(p2a_envs.app_id, 'id')
        self.assertEqual(p2a_envs.app_key, 'key')

    @requests_mock.Mocker()
    def test_get_advocates(self, m):

        m.get(self.p2a.uri + 'advocates', json=adv_json)

        adv_exp = ['id', 'prefix', 'firstname', 'middlename',
                   'lastname', 'suffix', 'notes', 'stage', 'connections',
                   'created_at_date', 'created_at_timezone',
                   'created_at_timezone_type', 'updated_at_date',
                   'updated_at_timezone', 'updated_at_timezone_type',
                   'address_city', 'address_county', 'address_latitude',
                   'address_longitude', 'address_state', 'address_street1',
                   'address_street2', 'address_zip4', 'address_zip5',
                   'districts_cityCouncil', 'districts_congressional',
                   'districts_stateHouse', 'districts_stateSenate']

        self.assertTrue(validate_list(adv_exp, self.p2a.get_advocates()['advocates']))
        ids_exp = ['advocate_id', 'ids']

        self.assertTrue(validate_list(ids_exp, self.p2a.get_advocates()['ids']))

        phone_exp = ['advocate_id', 'phones_address', 'phones_id', 'phones_subscribed']
        self.assertTrue(validate_list(phone_exp, self.p2a.get_advocates()['phones']))

        tags_exp = ['advocate_id', 'tags']
        self.assertTrue(validate_list(tags_exp, self.p2a.get_advocates()['tags']))

        email_exp = ['advocate_id', 'emails_address', 'emails_id', 'emails_subscribed']
        self.assertTrue(validate_list(email_exp, self.p2a.get_advocates()['emails']))

        member_exp = ['advocate_id', 'memberships_campaignid', 'memberships_created_at',
                      'memberships_id', 'memberships_name', 'memberships_source']
        self.assertTrue(validate_list(member_exp, self.p2a.get_advocates()['memberships']))

        fields_exp = ['advocate_id', 'fields']
        self.assertTrue(validate_list(fields_exp, self.p2a.get_advocates()['fields']))

    @requests_mock.Mocker()
    def test_get_campaigns(self, m):

        camp_exp = ['id', 'name', 'display_name', 'subtitle',
                    'public', 'topic', 'type', 'link', 'restrict_allow',
                    'updated_at_date', 'updated_at_timezone',
                    'updated_at_timezone_type', 'content_background_image',
                    'content_call_to_action', 'content_introduction',
                    'content_summary', 'content_thank_you']

        m.get(self.p2a.uri + 'campaigns', json=camp_json)

        self.assertTrue(validate_list(camp_exp, self.p2a.get_campaigns()))



import logging

logger = logging.getLogger(__name__)

"""
The BaseTable class is a child class for generating a table classes that are
specific to each flavor of SQL database.
"""


class BaseTable:
    """
    Base Table class object.
    """

    def __init__(self, database_connection, table_name):

        self.table = table_name
        self.db = database_connection
        self._columns = None

    @property
    def num_rows(self):
        """
        Get the number of rows in the table.
        """

        return self.db.query(f"SELECT COUNT(*) FROM {self.table}").first

    def max_primary_key(self, primary_key):
        """
        Get the maximum primary key in the table.
        """

        return self.db.query(f"SELECT MAX({primary_key}) FROM {self.table}").first

    def distinct_primary_key(self, primary_key):
        """
        Check if the passed primary key column is distinct.
        """

        sql = f"""
               SELECT
               COUNT(*) - COUNT(DISTINCT {primary_key})
               FROM {self.table}
               """

        if self.db.query(sql).first > 0:
            return False
        else:
            return True

    @property
    def columns(self):
        """
        Return a list of columns in the table.
        """

        if not self._columns:
            sql = f"SELECT * FROM {self.table} LIMIT 1"
            self._columns = self.db.query(sql).columns

        return self._columns

    @property
    def exists(self):
        """
        Check if table exists.
        """

        return self.db.table_exists(self.table)

    def get_rows(self, offset=0, chunk_size=None):
        """
        Get rows from a table.
        """

        sql = f"SELECT * FROM {self.table}"

        if chunk_size:
            sql += f" LIMIT {chunk_size}"

        sql += " OFFSET {offset}"

        return self.db.query(sql)

    def get_new_rows_count(self, primary_key_col, start_value):
        """
        Get a count of rows that have a greater primary key value
        than the one provided.
        """

        sql = f"""
               SELECT
               COUNT(*)
               FROM {self.table}
               WHERE {primary_key_col} > %s
               """

        return self.db.query(sql, [start_value]).first

    def get_new_rows(self, primary_key, cutoff_value, offset=0, chunk_size=None):
        """
        Get rows that have a greater primary key value than the one
        provided.

        It will select every value greater than the provided value.
        """

        if cutoff_value is not None:
            where_clause = f'where {primary_key} > %s'
            parameters = [cutoff_value]
        else:
            where_clause = ''
            parameters = []

        sql = f"""
               SELECT
               *
               FROM {self.table}
               {where_clause}
               """

        if chunk_size:
            sql += f" LIMIT {chunk_size}"

        sql += f" OFFSET {offset}"

        return self.db.query(sql, parameters)

    def drop(self, cascade=False):
        """
        Drop the table.
        """

        sql = f'DROP TABLE {self.table}'
        if cascade:
            sql += ' CASCADE'

        self.db.query(sql)
        logger.info(f'{self.table} dropped.')

    def truncate(self):
        """
        Truncate the table.
        """

        self.db.query(f'TRUNCATE TABLE {self.table}')
        logger.info(f'{self.table} truncated.')



"""NGPVAN Score Endpoints"""

from parsons.etl.table import Table
from parsons.utilities import cloud_storage
import uuid
import logging
import petl

logger = logging.getLogger(__name__)


class Scores(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_scores(self):
        """
        Get all scores.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('scores'))
        logger.info(f'Found {tbl.num_rows} scores.')
        return tbl

    def get_score(self, score_id):
        """
        Get an individual score.

        `Args:`
            score_id: int
                The score id
        `Returns:`
            dict
        """

        r = self.connection.get_request(f'scores/{score_id}')
        logger.info(f'Found score {score_id}.')
        return r

    def get_score_updates(self, created_before=None, created_after=None, score_id=None):
        """
        Get score updates.

        `Args:`
            created_before: str
                Filter score updates to those created before date. Use "YYYY-MM-DD"
                format.
            created_after: str
                Filter score updates to those created after date. Use "YYYY-MM-DD"
                format.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {'createdBefore': created_before,
                  'createdAfter': created_after,
                  'scoreId': score_id}

        tbl = Table(self.connection.get_request('scoreUpdates', params=params))
        if tbl.num_rows:
            tbl.unpack_dict('updateStatistics', prepend=False)
            tbl.unpack_dict('score', prepend=False)
        logger.info(f'Found {tbl.num_rows} score updates.')
        return tbl

    def get_score_update(self, score_update_id):
        """
        Get a score update object

            `Args:`
                score_update_id : int
                        The score update id
            `Returns:`
                dict
        """

        r = self.connection.get_request(f'scoreUpdates/{score_update_id}')
        logger.info(f'Returning score update {score_update_id}.')
        return r

    def update_score_status(self, score_update_id, status):
        """
        Change the status of a score update object. This end point is used to
        approve a score loading job.

        `Args:`
            score_update_id: str
                The score update id
            status: str
                One of 'pending approval', 'approved', 'disapproved'
        `Returns:`
            ``None``
        """

        if status not in ['pending approval', 'approved', 'disapproved',
                          'canceled']:

            raise ValueError("""Valid inputs for status are, 'pending approval',
                             'approved','disapproved','canceled'""")

        else:
            if status == 'pending approval':
                status = 'PendingApproval'
            else:
                status = status.capitalize()

        json = {"loadStatus": status}

        r = self.connection.patch_request(f'scoreUpdates/{score_update_id}', json=json)
        logger.info(f'Score {score_update_id} status updated to {status}.')
        return r

    def upload_scores(self, tbl, config, url_type, id_type='vanid', email=None, auto_approve=True,
                      approve_tolerance=.1, **url_kwargs):
        """
        Upload scores. Use to create or overwrite scores. Multiple score loads
        should be configured in a single call. [1]_

        `Args:`
            tbl: object
                A parsons.Table object. The table must contain the scores and first column in the
                table must contain the primary key (e.g. vanid).
            config: list
                The score configuration. A list of dictionaries in which you specify the following

                .. list-table::
                    :widths: 20 80
                    :header-rows: 0

                    * - ``score_column``
                      - The name of the column where the score is housed.
                    * - ``score_id``
                      - The score slot id.

                Example:

                .. highlight:: python
                .. code-block:: python

                  [{'score1_id' : int, score1_column': str}
                   {'score2_id' : int, score2_column': str}]

            url_type: str
                The cloud file storage to use to post the file. Currently only ``S3``.
            email: str
                An email address to send job load status updates.
            auto_approve: boolean
                If the scores are within the expected tolerance of deviation from the
                average values provided, then score will be automatically approved.
            approve_tolderance: float
                The deviation from the average scores allowed in order to automatically
                approve the score. Maximum of .1.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type.
                    * S3 requires ``bucket`` argument and, if not stored as env variables
                      ``aws_access_key`` and ``aws_secret_access_key``.
        `Returns:`
            int
               The score load job id.

        .. [1] NGPVAN asks that you load multiple scores in a single call to reduce the load
           on their servers.
        """

        # Move to cloud storage
        file_name = str(uuid.uuid1())
        url = cloud_storage.post_file(tbl, url_type, file_path=file_name + '.zip', **url_kwargs)
        logger.info(f'Table uploaded to {url_type}.')

        # Generate shell request
        json = {"description": 'A description',
                "file": {
                    "columnDelimiter": 'csv',
                    "columns": [{'name': c} for c in tbl.columns],
                    "fileName": file_name + '.csv',
                    "hasHeader": "True",
                    "hasQuotes": "False",
                    "sourceUrl": url},
                "actions": []
                }

        # Configure each score
        for i in config:
            action = {"actionType": "score",
                      "personIdColumn": tbl.columns[0],
                      "personIdType": id_type,
                      "scoreColumn": i['score_column'],
                      "scoreId": i['score_id']}

            if auto_approve:
                average = petl.stats(tbl.table, i['score_column']).mean
                action['approvalCriteria'] = {"average": average, "tolerance": approve_tolerance}

            json['actions'].append(action)

        # Add email listener
        if email:
            json['listeners'] = [{"type": "EMAIL", 'value': email}]

        # Upload scores
        r = self.connection.post_request('fileLoadingJobs', json=json)
        logger.info(f"Scores job {r['jobId']} created.")
        return r['jobId']


class FileLoadingJobs(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def create_file_load(self, file_name, file_url, columns, id_column, id_type,
                         score_id, score_column, delimiter='csv', header=True, quotes=True,
                         description=None, email=None, auto_average=None,
                         auto_tolerance=None):
        """
        .. warning::
           .. deprecated:: 0.7 Use :func:`parsons.VAN.upload_scores` instead.

        Loads a file. Only used for loading scores at this time. Scores must be
        compressed using `zip`.

        `Args:`
            file_name: str
                The name of the file contained in the zip file.
            file_url: str
                The url path to directly download the file. Can also be a path to an FTP site.
            columns: list
                A list of column names contained in the file.
            id_column: str
                The column name of the id column in the file.
            id_type: str
                A valid primary key, such as `VANID` or `DWID`. Varies by VAN instance.
            score_id: int
                The score slot id
            score_column: str
                    The column holding the score
            delimiter: str
                    The file delimiter used.
            email: str
                A valid email address in which file loading status will be sent.
            auto_average: float
                The average of scores to be loaded.
            auto_tolerance: float
                The fault tolerance of the VAN calculated average compared to the ``auto_average``.
                The tolerance must be less than 10% of the difference between the maximum and
                minimum possible acceptable values of the score.
        `Returns:`
            dict
                The file load id
        """

        columns = [{'name': c} for c in columns]

        # To Do: Validate that it is a .zip file. Not entirely sure if this is possible
        # as some urls might not end in ".zip".

        if delimiter not in ['csv', 'tab', 'pipe']:
            raise ValueError("Delimiter must be one of 'csv', 'tab' or 'pipe'")

        delimiter = delimiter.capitalize()

        json = {"description": 'A description',
                "file": {
                    "columnDelimiter": delimiter,
                    "columns": columns,
                    "fileName": file_name,
                    "hasHeader": header,
                    "hasQuotes": quotes,
                    "sourceUrl": file_url
                },
                "actions": [
                    {"actionType": "score",
                     "personIdColumn": id_column,
                     "personIdType": id_type,
                     "scoreColumn": score_column,
                     "scoreId": score_id}],
                "listeners": [
                    {"type": "EMAIL",
                     "value": email}]
                }

        if auto_average and auto_tolerance:

            json["actions"]["approvalCriteria"] = {"average": auto_average,
                                                   "tolerance": auto_tolerance}

        r = self.connection.post_request('fileLoadingJobs', json=json)['jobId']
        logger.info(f'Score loading job {r} created.')
        return r

    def create_file_load_multi(self, file_name, file_url, columns, id_column, id_type,
                               score_map, delimiter='csv', header=True, quotes=True,
                               description=None, email=None):
        """
        .. warning::
           .. deprecated:: 0.7 Use :func:`parsons.VAN.upload_scores` instead.

        An iteration of the :meth:`file_load` method that allows you to load multiple scores
        at the same time.

        `Args:`
            file_name : str
                The name of the file contained in the zip file.
            file_url : str
                The url path to directly download the file. Can also be a path to an FTP site.
            columns: list
                A list of column names contained in the file.
            id_column : str
                The column name of the id column in the file.
            id_type : str
                A valid primary key, such as `VANID` or `DWID`. Varies by VAN instance.
            score_map : list
                A list of dicts that adheres to the following syntax

                .. highlight:: python
                .. code-block:: python

                    [{'score_id' : int,
                      'score_column': str,
                      'auto_average': float,
                      'auto_tolerance': float }]

            email: str
                A valid email address in which file loading status will be sent.
        `Returns:`
            The file load job id
        """

        columns = [{'name': c} for c in columns]

        # To Do: Validate that it is a .zip file. Not entirely sure if this is possible
        # as some urls might not end in ".zip".

        if delimiter not in ['csv', 'tab', 'pipe']:
            raise ValueError("Delimiter must be one of 'csv', 'tab' or 'pipe'")

        delimiter = delimiter.capitalize()

        json = {"description": 'A description',
                "file": {
                    "columnDelimiter": delimiter,
                    "columns": columns,
                    "fileName": file_name,
                    "hasHeader": header,
                    "hasQuotes": quotes,
                    "sourceUrl": file_url
                },
                "listeners": [
                    {"type": "EMAIL",
                     "value": email}]
                }

        actions = []

        for score in score_map:

            action = {"actionType": "score",
                      "personIdColumn": id_column,
                                    "personIdType": id_type,
                                    "scoreColumn": score['score_column'],
                                    "scoreId": score['score_id'],
                                    "approvalCriteria": {
                                        "average": score['auto_average'],
                                        "tolerance": score['auto_tolerance']
                                    }
                      }

            actions.append(action)

        json['actions'] = actions

        r = self.connection.post_request('fileLoadingJobs', json=json)['jobId']
        logger.info(f'Score loading job {r} created.')
        return r



import os
import unittest
import requests_mock
from parsons.ngpvan.van import VAN
from parsons.etl.table import Table
from test.utils import validate_list, assert_matching_tables
from requests.exceptions import HTTPError


class TestNGPVAN(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters", raise_for_status=False)

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_canvass_responses_contact_types(self, m):

        json = {"name": "Auto Dial",
                "contactTypeId": 19,
                "channelTypeName": "Phone"}

        m.get(self.van.connection.uri + 'canvassResponses/contactTypes', json=json)

        assert_matching_tables(Table(json), self.van.get_canvass_responses_contact_types())

    @requests_mock.Mocker()
    def test_get_canvass_responses_input_types(self, m):

        json = {"inputTypeId": 11, "name": "API"}
        m.get(self.van.connection.uri + 'canvassResponses/inputTypes', json=json)
        assert_matching_tables(Table(json), self.van.get_canvass_responses_input_types())

    @requests_mock.Mocker()
    def test_get_canvass_responses_result_codes(self, m):

        json = {
            "shortName": "BZ",
            "resultCodeId": 18,
            "name": "Busy",
            "mediumName": "Busy"
        }

        m.get(self.van.connection.uri + 'canvassResponses/resultCodes', json=json)
        assert_matching_tables(Table(json), self.van.get_canvass_responses_result_codes())

    @requests_mock.Mocker()
    def test_get_survey_questions(self, m):

        json = {u'count': 67, u'items': [{
            "status": "Active",
            "responses": [
                {"shortName": "1",
                 "surveyResponseId": 1288926,
                 "name": "1-Strong Walz",
                         "mediumName": "1"},
                {"shortName": "2",
                 "surveyResponseId": 1288928,
                 "name": "2-Lean Walz",
                         "mediumName": "2"}],
            "scriptQuestion": "Who do you support for Governor?",
            "name": "MN Governor Gen",
                    "surveyQuestionId": 311838,
                    "mediumName": "MNGovG",
                    "shortName": "MGG",
                    "type": "Candidate",
                    "cycle": 2018
        }],
            u'nextPageLink': None}

        m.get(self.van.connection.uri + 'surveyQuestions', json=json)

        expected = ['status', 'responses', 'scriptQuestion', 'name',
                    'surveyQuestionId', 'mediumName', 'shortName',
                    'type', 'cycle']

        self.assertTrue(validate_list(expected, self.van.get_survey_questions()))

    @requests_mock.Mocker()
    def test_get_supporter_groups(self, m):

        json = {"items": [
            {
                "id": 12,
                "name": "tmc",
                "description": "A fun group."
            },
            {
                "id": 13,
                "name": "tmc",
                "description": "A fun group."
            },
        ],
            "nextPageLink": None,
            "count": 3
        }

        m.get(self.van.connection.uri + 'supporterGroups', json=json)

        expected = ['id', 'name', 'description']

        self.van.get_supporter_groups()

    @requests_mock.Mocker()
    def test_get_supporter_group(self, m):

        json = {"id": 12, "name": "tmc", "description": "A fun group."}
        m.get(self.van.connection.uri + 'supporterGroups/12', json=json)

        # Test that columns are expected columns
        self.assertEqual(self.van.get_supporter_group(12), json)

    @requests_mock.Mocker()
    def test_delete_supporter_group(self, m):

        # Test good input
        good_supporter_group_id = 5
        good_ep = f'supporterGroups/{good_supporter_group_id}'
        m.delete(self.van.connection.uri + good_ep, status_code=204)
        self.van.delete_supporter_group(good_supporter_group_id)

        # Test bad input raises
        bad_supporter_group_id = 999
        bad_vanid = 99999
        bad_ep = f'supporterGroups/{bad_supporter_group_id}'
        m.delete(self.van.connection.uri + bad_ep, status_code=404)
        self.assertRaises(HTTPError, self.van.delete_supporter_group, bad_supporter_group_id)

    @requests_mock.Mocker()
    def test_add_person_supporter_group(self, m):

        # Test good input
        good_supporter_group_id = 5
        good_vanid = 12345
        good_uri = f'supporterGroups/{good_vanid}/people/{good_supporter_group_id}'
        m.put(self.van.connection.uri + good_uri, status_code=204)
        self.van.add_person_supporter_group(good_vanid, good_supporter_group_id)

        # Test bad input
        bad_supporter_group_id = 999
        bad_vanid = 99999
        bad_uri = f'supporterGroups/{bad_vanid}/people/{bad_supporter_group_id}'
        m.put(self.van.connection.uri + bad_uri, status_code=404)
        self.assertRaises(HTTPError, self.van.add_person_supporter_group, bad_vanid, bad_supporter_group_id)

    @requests_mock.Mocker()
    def test_delete_person_supporter_group(self, m):

        # Test good input
        good_supporter_group_id = 5
        good_vanid = 12345
        good_ep = f'supporterGroups/{good_vanid}/people/{good_supporter_group_id}'
        m.delete(self.van.connection.uri + good_ep, status_code=204)
        self.van.delete_person_supporter_group(good_vanid, good_supporter_group_id)

        # Test bad input raises
        bad_supporter_group_id = 999
        bad_vanid = 99999
        bad_ep = f'supporterGroups/{bad_vanid}/people/{bad_supporter_group_id}'
        m.delete(self.van.connection.uri + bad_ep, status_code=404)
        self.assertRaises(HTTPError, self.van.delete_person_supporter_group, bad_vanid, bad_supporter_group_id)

if __name__ == '__main__':

    unittest.main()



import json
from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector
import logging

logger = logging.getLogger(__name__)


class ActionNetwork(object):
    """
    `Args:`
        api_token: str
            The OSDI API token
        api_url:
            The end point url
    """
    def __init__(self, api_token=None, api_url=None):
        self.api_token = check_env.check('AN_API_TOKEN', api_token)
        self.headers = {
            "Content-Type": "application/json",
            "OSDI-API-Token": self.api_token
        }
        self.api_url = check_env.check('AN_API_URL', api_url)
        self.api = APIConnector(self.api_url, headers=self.headers)

    def _get_page(self, object_name, page, per_page=25):
        # returns data from one page of results
        if per_page > 25:
            per_page = 25
            logger.info("Action Network's API will not return more than 25 entries per page. \
            Changing per_page parameter to 25.")
        page_url = f"{self.api_url}/{object_name}?page={page}&per_page={per_page}"
        return self.api.get_request(url=page_url)

    def _get_entry_list(self, object_name, limit=None, per_page=25):
        # returns a list of entries for a given object, such as people, tags, or actions
        count = 0
        page = 1
        return_list = []
        while True:
            response = self._get_page(object_name, page, per_page)
            page = page + 1
            response_list = response['_embedded'][f"osdi:{object_name}"]
            if not response_list:
                return Table(return_list)
            return_list.extend(response_list)
            count = count + len(response_list)
            if limit:
                if count >= limit:
                    return Table(return_list[0:limit])

    def get_people(self, limit=None, per_page=25, page=None):
        """
        `Args:`
            limit:
                The number of entries to return. When None, returns all entries.
            per_page
                The number of entries per page to return. 25 maximum.
            page
                Which page of results to return
        `Returns:`
            A list of JSONs of people stored in Action Network.
        """
        if page:
            self._get_page("people", page, per_page)
        return self._get_entry_list("people", limit, per_page)

    def get_person(self, person_id):
        """
        `Args:`
            person_id:
                Id of the person.
        `Returns:`
            A  JSON of the entry. If the entry doesn't exist, Action Network returns
            "{'error': 'Couldn't find person with id = <id>'}"
        """
        return self.api.get_request(url=f"{self.api_url}/people/{person_id}")

    def add_person(self, email_address, given_name=None, family_name=None, tags=[],
                   languages_spoken=[], postal_addresses=[],
                   **kwargs):
        """
        `Args:`
            email_address:
                Can be any of the following
                    - a string with the person's email
                    - a dictionary with the following fields
                        - email_address (REQUIRED)
                        - primary (OPTIONAL): Boolean indicating the user's primary email address
                        - status (OPTIONAL): can taken on any of these values
                            - "subscribed"
                            - "unsubscribed"
                            - "bouncing"
                            - "previous bounce"
                            - "spam complaint"
                            - "previous spam complaint"
            given_name:
                The person's given name
            family_name:
                The person's family name
            tags:
                Any tags to be applied to the person
            languages_spoken:
                Optional field. A list of strings of the languages spoken by the person
            postal_addresses:
                Optional field. A list of dictionaries.
                For details, see Action Network's documentation:
                https://actionnetwork.org/docs/v2/people#put
            **kwargs:
                Any additional fields to store about the person. Action Network allows
                any custom field.
        Adds a person to Action Network
        """
        email_addresses_field = None
        if type(email_address) == str:
            email_addresses_field = [{"address": email_address}]
        elif type(email_address == list):
            if type(email_address[0]) == str:
                email_addresses_field = [{"address": email} for email in email_address]
                email_addresses_field[0]['primary'] = True
            if type(email_address[0]) == dict:
                email_addresses_field = email_address
        if not email_addresses_field:
            raise("email_address must be a string, list of strings, or list of dictionaries")
        data = {
            "person": {
                "email_addresses": email_addresses_field,
                "given_name": given_name,
                "family_name": family_name,
                "languages_spoken": languages_spoken,
                "postal_addresses": postal_addresses,
                "custom_fields": {**kwargs}
              },
            "add_tags": tags
        }
        response = self.api.post_request(url=f"{self.api_url}/people", data=json.dumps(data))
        identifiers = response['identifiers']
        person_id = [entry_id.split(':')[1]
                     for entry_id in identifiers if 'action_network:' in entry_id][0]
        logger.info(f"Entry {person_id} successfully added to people.")
        return response

    def update_person(self, entry_id, **kwargs):
        """
        `Args:`
            entry_id:
                The person's Action Network id
            **kwargs:
                Fields to be updated. The possible fields are
                    email_address:
                        Can be any of the following
                            - a string with the person's email
                            - a dictionary with the following fields
                                - email_address (REQUIRED)
                                    - primary (OPTIONAL): Boolean indicating the user's
                                    primary email address
                                - status (OPTIONAL): can taken on any of these values
                                    - "subscribed"
                                    - "unsubscribed"
                                    - "bouncing"
                                    - "previous bounce"
                                    - "spam complaint"
                                    - "previous spam complaint"
                    given_name:
                        The person's given name
                    family_name:
                        The person's family name
                    tags:
                        Any tags to be applied to the person
                    languages_spoken:
                        Optional field. A list of strings of the languages spoken by the person
                    postal_addresses:
                        Optional field. A list of dictionaries.
                        For details, see Action Network's documentation:
                        https://actionnetwork.org/docs/v2/people#put
                    custom_fields:
                        A dictionary of any other fields to store about the person.
        Updates a person's data in Action Network
        """
        data = {**kwargs}
        response = self.api.put_request(url=f"{self.api_url}/people/{entry_id}",
                                        json=json.dumps(data), success_codes=[204, 201, 200])
        logger.info(f"Person {entry_id} successfully updated")
        return response

    def get_tags(self, limit=None, per_page=25, page=None):
        """
        `Args:`
            limit:
                The number of entries to return. When None, returns all entries.
            per_page
                The number of entries per page to return. 25 maximum.
            page
                Which page of results to return
        `Returns:`
            A list of JSONs of tags in Action Network.
        """
        if page:
            self.get_page("tags", page, per_page)
        return self._get_entry_list("tags", limit, per_page)

    def get_tag(self, tag_id):
        """
        `Args:`
            tag_id:
                Id of the tag.
        `Returns:`
            A  JSON of the entry. If the entry doesn't exist, Action Network returns
            "{'error': 'Couldn't find tag with id = <id>'}"
        """
        return self.api.get_request(url=f"{self.api_url}/tags/{tag_id}")

    def add_tag(self, name):
        """
        `Args:`
            name:
                The tag's name. This is the ONLY editable field
        Adds a tag to Action Network. Once created, tags CANNOT be edited or deleted.
        """
        data = {
            "name": name
        }
        response = self.api.post_request(url=f"{self.api_url}/tags", data=json.dumps(data))
        identifiers = response['identifiers']
        person_id = [entry_id.split(':')[1]
                     for entry_id in identifiers if 'action_network:' in entry_id][0]
        logger.info(f"Tag {person_id} successfully added to tags.")
        return response



from parsons.etl.table import Table
import logging

logger = logging.getLogger(__name__)


class CustomFields():

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_custom_fields(self, field_type='contacts'):
        """
        Get custom fields.

        `Args:`
            field_type : str
                Filter by custom field group type. Must be one of ``contacts`` or
                ``contributions``.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {'customFieldsGroupType': field_type.capitalize()}

        tbl = Table(self.connection.get_request('customFields', params=params))
        logger.info(f'Found {tbl.num_rows} custom fields.')
        return tbl

    def get_custom_fields_values(self, field_type='contacts'):
        """
        Get custom field values as a long table.

        `Args:`
            field_type : str
                Filter by custom field group type. Must be one of ``contacts`` or
                ``contributions``.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self.get_custom_fields()

        # Some custom fields do no have associated values. If this is the case then
        # we should return an empty Table, but with the expected columns.
        if tbl.get_column_types('availableValues') == ['NoneType']:
            logger.info(f'Found 0 custom field values.')
            return Table([{'customFieldId': None,
                           'id': None,
                           'name': None,
                           'parentValueId': None}])

        else:
            logger.info(f'Found {tbl.num_rows} custom field values.')
            return tbl.long_table('customFieldId', 'availableValues', prepend=False)

    def get_custom_field(self, custom_field_id):
        """
        Get a custom field.

        `Args:`
            custom_field_id: int
                A valid custom field id.
        `Returns:`
            A json.
        """

        r = self.connection.get_request(f'customFields/{custom_field_id}')
        logger.info(f'Found custom field {custom_field_id}.')
        return r



import importlib
import inspect
import json
import os

import boto3


"""

In lambda handler:

from parsons.aws import event_command

def handler(event, context):

    ## ADD THESE TWO LINES TO TOP OF HANDLER:
    if event_command(event, context):
         return

"""
try:
    from zappa.asynchronous import run as zappa_run
except ImportError:
    zappa_run = None


def event_command(event, context):
    """
    Minimal shim to add to the top lambda handler function
    to enable distributed tasks
    In your lambda handler:

    .. code-block:: python
       :emphasize-lines: 5,6

       from parsons.aws import event_command

       def handler(event, context):
           ## ADD THESE TWO LINES TO TOP OF HANDLER:
           if event_command(event, context):
               return

    The rest of this library is compatible with zappa.async library.
    If you have deployed your app with `Zappa <https://github.com/Miserlou/Zappa>`,
    then you do NOT need to add this shim.
    """
    if not set(event).intersection({'task_path', 'args', 'kwargs'}):
        return False  # did not match an event command
    func = import_and_get_task(event['task_path'], event.get('func_class_init_kwargs'))
    # if the func was decorated with zappa.async.task then run the real function
    func = getattr(func, 'sync', func)

    # DID match an event command
    # -- so probably don't do the usual thing the Lambda handler does
    return (func(*event['args'], **event['kwargs'])
            or True)


def run(func, args=[], kwargs={}, service='lambda', capture_response=False,
        remote_aws_lambda_function_name=None, remote_aws_region=None,
        func_class=None, func_class_init_kwargs=None, **task_kwargs):
    lambda_function_name = (remote_aws_lambda_function_name
                            or os.environ.get('AWS_LAMBDA_FUNCTION_NAME'))
    if not lambda_function_name or lambda_function_name == 'FORCE_LOCAL':
        # We are neither running in Lambda environment, nor given one to invoke
        # so let's run it synchronously -- so code can be compatible both in-and-out of Lambda
        func(*args, **kwargs)
        return True
    # zappa has more robust and allows more configs -- but is not compatible with func_class
    if zappa_run and not func_class:
        return zappa_run(func, args, kwargs, service,
                         capture_response, remote_aws_lambda_function_name, remote_aws_region,
                         **task_kwargs)

    task_path = get_func_task_path(func, func_class)
    payload = (json.dumps({'task_path': task_path,
                           'args': args,
                           'kwargs': kwargs,
                           'func_class_init_kwargs': func_class_init_kwargs})
               .encode('utf-8'))
    if len(payload) > 128000:  # pragma: no cover
        raise AsyncException("Payload too large for async Lambda call")
    lambda_client = boto3.Session().client('lambda')
    response = lambda_client.invoke(
        FunctionName=lambda_function_name,
        InvocationType='Event',  # makes the call async
        Payload=payload
    )
    return response.get('StatusCode', 0) == 202


##
# Utility Functions
##

def import_and_get_task(task_path, instance_init_kwargs=None):
    """
    Given a modular path to a function, import that module
    and return the function.
    """
    module, function = task_path.rsplit('.', 1)
    app_module = importlib.import_module(module)
    class_func = function.split('|')
    app_function = getattr(app_module, class_func[0])
    if len(class_func) == 1:
        return app_function

    def init_and_run(*args, **kwargs):
        print('INITRUN', args, kwargs)
        if len(class_func) == 3:  # instance
            instance = app_function  # actually the class
        else:
            instance = app_function(**(instance_init_kwargs or {}))
        method = getattr(instance, class_func[1])
        return method(*args, **kwargs)
    return init_and_run


def get_func_task_path(func, method_class=None):
    """
    Format the modular task path for a function via inspection.
    """
    module_path = inspect.getmodule(method_class or func).__name__
    func_name = func.__name__

    # To support class methods, we need to see if it IS a method on a class
    # and then also determine if it is an instance method or a classmethod
    # Then we record that info with |'s to be decoded in import_and_get_task
    # classmethod format: "Foo|method|"
    # instance method format: "Foo|method"
    task_path = '{}.{}{}{}'.format(
        module_path,
        f'{method_class.__name__}|' if method_class else '',
        func_name,
        '|' if method_class and 'of <class' in repr(func) else ''
    )
    return task_path


class AsyncException(Exception):
    pass



from parsons.etl import Table
from parsons.notifications.slack import Slack

from slackclient.exceptions import SlackClientError

import os
import requests_mock
import unittest
import json


_dir = os.path.dirname(__file__)
responses_dir = f"{_dir}/responses"


class TestSlack(unittest.TestCase):

    def setUp(self):
        os.environ["SLACK_API_TOKEN"] = 'SOME_API_TOKEN'

        self.slack = Slack()

    def tearDown(self):

        pass

    def test_slack_init(self):
        # Delete to test that is raises an error
        del os.environ["SLACK_API_TOKEN"]

        self.assertNotIn("SLACK_API_TOKEN", os.environ)

        self.assertRaises(KeyError, Slack)

        os.environ["SLACK_API_TOKEN"] = 'SOME_API_TOKEN'
        self.assertIn("SLACK_API_TOKEN", os.environ)

    @requests_mock.Mocker()
    def test_channels(self, m):

        with open(f"{responses_dir}/channels.json", "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/conversations.list", json=slack_resp)

        tbl = self.slack.channels()

        self.assertIsInstance(tbl, Table)

        expected_columns = ["id", "name"]
        self.assertListEqual(tbl.columns, expected_columns)
        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_channels_all_fields(self, m):

        with open(f"{responses_dir}/channels.json", "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/conversations.list", json=slack_resp)

        fields_req = [
            "id", "name", "is_channel", "created", "creator",
            "is_archived", "is_general", "name_normalized", "is_shared",
            "is_org_shared", "is_member", "is_private", "is_mpim", "members",
            "topic_value", "topic_creator", "topic_last_set", "purpose_value",
            "purpose_creator", "purpose_last_set", "previous_names",
            "num_members"]
        tbl = self.slack.channels(fields=fields_req)

        self.assertIsInstance(tbl, Table)

        expected_columns = [
            "id", "name", "is_channel", "created", "creator",
            "is_archived", "is_general", "name_normalized", "is_shared",
            "is_org_shared", "is_member", "is_private", "is_mpim", "members",
            "topic_value", "topic_creator", "topic_last_set", "purpose_value",
            "purpose_creator", "purpose_last_set", "previous_names",
            "num_members"]

        self.assertListEqual(sorted(tbl.columns), sorted(expected_columns))
        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_users(self, m):

        with open(f"{responses_dir}/users.json", "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/users.list", json=slack_resp)

        tbl = self.slack.users()

        self.assertIsInstance(tbl, Table)

        expected_columns = ["id", "name", "deleted", "profile_email", "profile_real_name_normalized"]
        self.assertListEqual(tbl.columns, expected_columns)
        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_users_all_fields(self, m):

        with open(f"{responses_dir}/users.json", "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/users.list", json=slack_resp)

        fields_req = [
            "id", "team_id", "name", "deleted", "color", "real_name", "tz",
            "tz_label", "tz_offset", "is_admin", "is_owner",
            "is_primary_owner", "is_restricted", "is_ultra_restricted",
            "is_bot", "updated", "is_app_user", "has_2fa",
            "profile_avatar_hash", "profile_display_name",
            "profile_display_name_normalized", "profile_email",
            "profile_first_name", "profile_image_1024", "profile_image_192",
            "profile_image_24", "profile_image_32", "profile_image_48",
            "profile_image_512", "profile_image_72", "profile_image_original",
            "profile_last_name", "profile_phone", "profile_real_name",
            "profile_real_name_normalized", "profile_skype",
            "profile_status_emoji", "profile_status_text", "profile_team",
            "profile_title"]
        tbl = self.slack.users(fields=fields_req)

        self.assertIsInstance(tbl, Table)

        expected_columns = [
            "id", "team_id", "name", "deleted", "color", "real_name", "tz",
            "tz_label", "tz_offset", "is_admin", "is_owner",
            "is_primary_owner", "is_restricted", "is_ultra_restricted",
            "is_bot", "updated", "is_app_user", "has_2fa",
            "profile_avatar_hash", "profile_display_name",
            "profile_display_name_normalized", "profile_email",
            "profile_first_name", "profile_image_1024", "profile_image_192",
            "profile_image_24", "profile_image_32", "profile_image_48",
            "profile_image_512", "profile_image_72", "profile_image_original",
            "profile_last_name", "profile_phone", "profile_real_name",
            "profile_real_name_normalized", "profile_skype",
            "profile_status_emoji", "profile_status_text", "profile_team",
            "profile_title"]
        self.assertListEqual(sorted(tbl.columns), sorted(expected_columns))
        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_message_channel(self, m):

        with open(f"{responses_dir}/message_channel.json", "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/chat.postMessage", json=slack_resp)

        dct = self.slack.message_channel(
            "C1H9RESGL", "Here's a message for you")

        self.assertIsInstance(dct, dict)
        self.assertListEqual(sorted(dct), sorted(slack_resp))

        m.post(
            "https://slack.com/api/chat.postMessage",
            json={"ok": False, "error": "invalid_auth"})

        self.assertRaises(
            SlackClientError,
            self.slack.message_channel,
            "FakeChannel", "Here's a message for you")

    @requests_mock.Mocker()
    def test_file_upload(self, m):

        file_path = f"{responses_dir}/file_upload.json"
        with open(file_path, "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/files.upload", json=slack_resp)

        dct = self.slack.upload_file(
            ["D0L4B9P0Q"], file_path)

        self.assertIsInstance(dct, dict)
        self.assertListEqual(sorted(dct), sorted(slack_resp))

        m.post(
            "https://slack.com/api/files.upload",
            json={"ok": False, "error": "invalid_auth"})

        self.assertRaises(
            SlackClientError,
            self.slack.upload_file,
            ["D0L4B9P0Q"], file_path)



import datetime
import decimal
import os
import unittest
import requests_mock
from test.utils import assert_matching_tables

from parsons.etl.table import Table
from parsons.braintree.braintree import Braintree

_dir = os.path.dirname(__file__)


class TestBraintree(unittest.TestCase):

    def setUp(self):
        self.braintree = Braintree(merchant_id='abcd1234abcd1234', public_key='abcd1234abcd1234',
                                   private_key='abcd1234abcd1234abcd1234abcd1234')

    @requests_mock.Mocker()
    def test_dispute_search(self, m):
        m.post('https://api.braintreegateway.com:443'
               '/merchants/abcd1234abcd1234/disputes/advanced_search?page=1',
               text=open(f'{_dir}/test_data/dispute_example.xml').read())
        table = self.braintree.get_disputes(start_date="2020-01-01", end_date="2020-01-02")

        self.assertEqual(len(table.table), 3)
        self.assertEqual(table[0]['id'], 'abcd1234abcd1234')
        self.assertEqual(table[1]['id'], 'ghjk6789ghjk6789')
        self.assertEqual(table[0]['transaction_id'], 'd9f876fg')
        self.assertEqual(table[1]['transaction_id'], '98df87fg')
        self.assertEqual(table[0]['reason'], 'transaction_amount_differs')
        self.assertEqual(table[1]['reason'], 'fraud')

    @requests_mock.Mocker()
    def test_transaction_search(self, m):
        m.post('https://api.braintreegateway.com:443'
               '/merchants/abcd1234abcd1234/transactions/advanced_search_ids',
               text="""
               <search-results>
                  <page-size type="integer">50</page-size>
                  <ids type="array"><item>1234abcd</item> <item>0987asdf</item> </ids>
               </search-results>
        """)
        table = self.braintree.get_transactions(disbursement_start_date="2020-01-01",
                                                disbursement_end_date="2020-01-02",
                                                just_ids=True)
        assert_matching_tables(table, Table([['id'], ['1234abcd'], ['0987asdf']]))
        m.post('https://api.braintreegateway.com:443'
               '/merchants/abcd1234abcd1234/transactions/advanced_search',
               text=open(f'{_dir}/test_data/transaction_example.xml').read())
        full_table = self.braintree.get_transactions(disbursement_start_date="2020-01-01",
                                                     disbursement_end_date="2020-01-02",
                                                     table_of_ids=table)
        self.assertEqual(len(table.table), 3)
        self.assertEqual(len(full_table.table), 3)
        self.assertEqual(table[0]['id'], '1234abcd')
        self.assertEqual(table[1]['id'], '0987asdf')
        self.assertEqual(len(table[0].keys()), 1)
        self.assertEqual(len(full_table[0].keys()), 64)

        self.assertEqual(full_table[0]['disbursement_date'], datetime.date(2019, 12, 30))
        self.assertEqual(full_table[0]['credit_card_bin'], '789234')
        self.assertEqual(full_table[0]['disbursement_success'], True)
        self.assertEqual(full_table[0]['amount'], decimal.Decimal('150.00'))

    def test_query_generation(self):
        query = self.braintree._get_query_objects(
            'transaction', **{'disbursement_date': {'between': ["2020-01-01", "2020-01-01"]}})
        self.assertEqual(query[0].name, 'disbursement_date')
        self.assertEqual(query[0].to_param(), {'min': '2020-01-01', 'max': '2020-01-01'})

        query = self.braintree._get_query_objects(
            'transaction', **{'merchant_account_id': {'in_list': ["abc123"]}})

        self.assertEqual(query[0].name, 'merchant_account_id')
        self.assertEqual(query[0].to_param(), ['abc123'])

        query = self.braintree._get_query_objects(
            'dispute', **{
                'merchant_account_id': {'in_list': ["abc123"]},
                'effective_date': {'between': ["2020-01-01", "2020-01-01"]}})
        self.assertEqual(query[0].name, 'merchant_account_id')
        self.assertEqual(query[1].name, 'effective_date')
        self.assertEqual(query[1].to_param(), {'min': '2020-01-01', 'max': '2020-01-01'})



import unittest
import os
import requests_mock
import unittest.mock as mock
from parsons.ngpvan.van import VAN
from parsons.etl.table import Table
from test.utils import validate_list, assert_matching_tables
from parsons.utilities import cloud_storage


os.environ['VAN_API_KEY'] = 'SOME_KEY'

class TestScores(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters", raise_for_status=False)

    @requests_mock.Mocker()
    def test_get_scores(self, m):

        json = {u'count': 2, u'items':
                [{u'origin': None,
                  u'scoreId': 2716,
                  u'name': u'Democratic Party Support',
                  u'maxValue': 100.0,
                  u'minValue': 1.0,
                  u'state': None,
                  u'shortName': u'Dem Support',
                  u'description': None}],
                u'nextPageLink': None}

        m.get(self.van.connection.uri + 'scores', json=json)

        expected = ['origin', 'scoreId', 'name', 'maxValue', 'minValue',
                    'state', 'shortName', 'description']

        self.assertTrue(validate_list(expected, self.van.get_scores()))

    @requests_mock.Mocker()
    def test_get_score(self, m):

        score_id = 2716

        json = {u'origin': None,
                u'scoreId': 2716,
                u'name': u'Democratic Party Support',
                u'maxValue': 100.0,
                u'minValue': 1.0,
                u'state': None,
                u'shortName': u'Dem Support',
                u'description': None}

        m.get(self.van.connection.uri + 'scores/{}'.format(score_id), json=json)
        self.assertEqual(json, self.van.get_score(score_id))

    @requests_mock.Mocker()
    def test_get_score_updates(self, m):

        json = {'items': [{
                'scoreUpdateId': 58319,
                'score': {
                  'scoreId': 29817,
                  'name': 'TargetSmart Gun Ownership',
                  'shortName': None,
                  'description': None,
                  'minValue': 0.0,
                  'maxValue': 100.0,
                  'state': 'MT',
                  'origin': None
                },
                'updateStatistics': {
                  'totalRows': 856644,
                  'duplicateRows': 0,
                  'matchedRows': 856644,
                  'matchPercent': 100.0,
                  'increasedBy': 441264,
                  'decreasedBy': 280588,
                  'nulledOut': 3649,
                  'added': 115129,
                  'outOfRange': 0,
                  'badValues': 0,
                  'maxValue': 95.9,
                  'minValue': 11.2,
                  'averageValue': 72.3338,
                  'medianValue': 76.3
                },
                'loadStatus': 'Completed',
                'dateProcessed': '2019-09-10T02:07:00Z'
              }],
              'nextPageLink': None,
              'count': 306
            }

        m.get(self.van.connection.uri + 'scoreUpdates', json=json)

        expected = ['scoreUpdateId', 'loadStatus', 'dateProcessed', 'added', 'averageValue',
                    'badValues', 'decreasedBy', 'duplicateRows', 'increasedBy', 'matchPercent',
                    'matchedRows', 'maxValue', 'medianValue', 'minValue', 'nulledOut',
                    'outOfRange', 'totalRows', 'description', 'maxValue', 'minValue', 'name',
                    'origin', 'scoreId', 'shortName', 'state']

        self.assertTrue(validate_list(expected, self.van.get_score_updates()))

    @requests_mock.Mocker()
    def test_get_score_update(self, m):

        score_update_id = 27892

        json = {
            "loadStatus": "Canceled",
            "updateStatistics": {
                "increasedBy": 1,
                "nulledOut": 1,
                "added": 0,
                "matchedRows": 4,
                "matchPercent": 100.0,
                "outOfRange": 0,
                "badValues": 1,
                "totalRows": 4,
                "maxValue": 30.0,
                "medianValue": 15.0,
                "minValue": 10.0,
                "duplicateRows": "null",
                "averageValue": 20.0,
                "decreasedBy": 2
            },
            "score": {
                "origin": "null",
                "scoreId": 2716,
                "name": "Democratic Party Support",
                        "maxValue": 100.0,
                        "minValue": 1.0,
                        "state": "null",
                        "shortName": "null",
                        "description": "null"
            },
            "dateProcessed": "null",
            "scoreUpdateId": 27892
        }

        m.get(self.van.connection.uri + f'scoreUpdates/{score_update_id}', json=json)

        expected = ['loadStatus', 'updateStatistics', 'score', 'dateProcessed', 'scoreUpdateId']

        self.assertEqual(json, self.van.get_score_update(score_update_id))

    @requests_mock.Mocker()
    def test_update_score_status(self, m):

        score_update_id = 27892

        m.patch(self.van.connection.uri + 'scoreUpdates/{}'.format(score_update_id),
                status_code=204)

        # Test bad input
        self.assertRaises(ValueError, self.van.update_score_status, score_update_id, 'not a thing.')

        # Test good input
        self.assertTrue(self.van.update_score_status(score_update_id, 'approved'))

    @requests_mock.Mocker()
    def test_upload_scores(self, m):

        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = 'https://box.com/my_file.zip'

        # Test uploading a job
        tbl = Table([['vanid', 'col'], ['1', '.5']])
        json = {'jobId': 9749}
        m.post(self.van.connection.uri + 'FileLoadingJobs', json=json, status_code=201)
        job_id = self.van.upload_scores(tbl, [{'score_id': 9999, 'score_column': 'col'}], url_type='S3')

    @requests_mock.Mocker()
    def test_create_file_load(self, m):

        file_name = 'test_scores.csv'
        file_url_good = 'http://tmc.org/test_scores.zip'
        file_url_bad = 'http://tmc.org/test_scores'
        columns = ['vanid', 'score']
        id_column = 'vanid'
        id_type = 'VANID'
        score_id = 2716
        score_column = 'score'
        bad_delimiter = '*'

        json = {'jobId': 9749}

        m.post(self.van.connection.uri + 'FileLoadingJobs', json=json, status_code=201)

        # Test bad delimiter
        self.assertRaises(ValueError, self.van.create_file_load, file_name, file_url_good, columns,
                          id_column, id_type, score_id, score_column, delimiter=bad_delimiter)

        # Test good request
        self.assertEqual(json['jobId'], self.van.create_file_load(file_name, file_url_good, columns,
                                                                  id_column, id_type, score_id,
                                                                  score_column))


from parsons.tools import credential_tools as ct
import json
import os
import shutil
import unittest


class TestCredentialTool(unittest.TestCase):

    def setUp(self):

        os.environ['TES_VAR1'] = 'variable1'
        os.environ['TES_VAR2'] = 'variable2'

        self.tmp_folder = "tmp"
        self.json_file = "credentials.json"
        os.mkdir(self.tmp_folder)

        with open(f"{self.tmp_folder}/{self.json_file}", 'w') as f:
            f.write(json.dumps({
                "json": "file"
            }))

    def tearDown(self):

        # Delete tmp folder and files
        shutil.rmtree(self.tmp_folder)

    def test_decode_credential(self):
        encoded_cred = ("PRSNSENVeyJFTkNfVkFSMSI6ICJlbmNvZGVkLXZhcmlhYmxl"
                        "LTEiLCAiRU5DX1ZBUjIiOiAiZW5jLXZhci0yIn0=")

        expected = {
            "ENC_VAR1": "encoded-variable-1",
            "ENC_VAR2": "enc-var-2"}

        self.assertDictEqual(
            ct.decode_credential(encoded_cred, export=False), expected)

    def test_decode_credential_export(self):
        encoded_cred = ("PRSNSENVeyJFTkNfVkFSMSI6ICJlbmNvZGVkLXZhcmlhYmxl"
                        "LTEiLCAiRU5DX1ZBUjIiOiAiZW5jLXZhci0yIn0=")

        expected = {
            "ENC_VAR1": "encoded-variable-1",
            "ENC_VAR2": "enc-var-2"}

        self.assertNotIn("ENC_VAR1", os.environ)
        self.assertNotIn("ENC_VAR2", os.environ)

        ct.decode_credential(encoded_cred)

        self.assertIn("ENC_VAR1", os.environ)
        self.assertIn("ENC_VAR2", os.environ)

        self.assertEqual(os.environ['ENC_VAR1'], expected['ENC_VAR1'])
        self.assertEqual(os.environ['ENC_VAR2'], expected['ENC_VAR2'])

    def test_decode_credential_save(self):
        encoded_cred = ("PRSNSENVeyJFTkNfVkFSMSI6ICJlbmNvZGVkLXZhcmlhYmxl"
                        "LTEiLCAiRU5DX1ZBUjIiOiAiZW5jLXZhci0yIn0=")

        expected = {
            "ENC_VAR1": "encoded-variable-1",
            "ENC_VAR2": "enc-var-2"}

        file_path = f"{self.tmp_folder}/saved_credentials.json"
        self.assertFalse(os.path.isfile(file_path))

        ct.decode_credential(encoded_cred, export=False, save_path=file_path)

        self.assertTrue(os.path.isfile(file_path))

        with open(file_path, 'r') as f:
            cred = json.load(f)

        self.assertDictEqual(cred, expected)

    def test_decode_credential_error(self):
        non_json = "non-json string"

        self.assertRaises(ValueError, ct.decode_credential, non_json)

    def test_encode_from_json_str(self):
        json_str = '{"json": "string"}'
        expected = "PRSNSENVeyJqc29uIjogInN0cmluZyJ9"

        self.assertEqual(ct.encode_from_json_str(json_str), expected)

    def test_encode_from_json_file(self):
        json_path = f"{self.tmp_folder}/{self.json_file}"
        expected = "PRSNSENVeyJqc29uIjogImZpbGUifQ=="

        self.assertEqual(
            ct.encode_from_json_file(json_path), expected)

    def testencode_from_env(self):
        lst = ['TES_VAR1', 'TES_VAR2']
        expected = ("PRSNSENVeyJURVNfVkFSMSI6ICJ2YXJpYWJsZTEiLCAiVEVTX1ZBU"
                    "jIiOiAidmFyaWFibGUyIn0=")

        self.assertEqual(ct.encode_from_env(lst), expected)

    def test_encode_from_dict(self):
        dct = {"dict": "variable"}
        expected = "PRSNSENVeyJkaWN0IjogInZhcmlhYmxlIn0="

        self.assertEqual(ct.encode_from_dict(dct), expected)


if __name__ == "__main__":
    unittest.main()



import json
import logging
import requests

from parsons.etl.table import Table
from parsons.utilities import check_env

logger = logging.getLogger(__name__)


class ActionKit(object):
    """
    Instatiate the ActionKit class

    `Args:`
        domain: str
            The ActionKit domain (e.g. ``myorg.actionkit.com``) Not required if
            ``ACTION_KIT_DOMAIN`` env variable set.
        username: str
            The authorized ActionKit username. Not required if ``ACTION_KIT_USERNAME`` env
            variable set.
        password: str
            The authorized ActionKit user password. Not required if ``ACTION_KIT_USERNAME``
            env variable set.
    """

    _default_headers = {'content-type': 'application/json',
                        'accepts': 'application/json'}

    def __init__(self, domain=None, username=None, password=None):

        self.domain = check_env.check('ACTION_KIT_DOMAIN', domain)
        self.username = check_env.check('ACTION_KIT_USERNAME', username)
        self.password = check_env.check('ACTION_KIT_PASSWORD', password)
        self.conn = self._conn()

    def _conn(self, default_headers=_default_headers):

        client = requests.Session()
        client.auth = (self.username, self.password)
        client.headers.update(default_headers)
        return client

    def _base_endpoint(self, endpoint, entity_id=None):
        # Create the base endpoint URL

        url = f'https://{self.domain}/rest/v1/{endpoint}/'

        if entity_id:
            return url + f'{entity_id}/'
        return url

    def _base_get(self, endpoint, entity_id=None, exception_message=None):
        # Make a general get request to ActionKit

        resp = self.conn.get(self._base_endpoint(endpoint, entity_id))

        if exception_message and resp.status_code == 404:
            raise Exception(self.parse_error(resp, exception_message))

        return resp.json()

    def _base_post(self, endpoint, exception_message, return_full_json=False, **kwargs):
        # Make a general post request to ActionKit

        resp = self.conn.post(self._base_endpoint(endpoint), data=json.dumps(kwargs))

        if resp.status_code != 201:
            raise Exception(self.parse_error(resp, exception_message))

        # Some of the methods should just return pointer to location of created
        # object.
        if 'headers' in resp.__dict__ and not return_full_json:
            return resp.__dict__['headers']['Location']

        # Not all responses return a json
        try:
            return resp.json()

        except ValueError:
            return None

    def parse_error(self, resp, exception_message):
        # AK provides some pretty robust/helpful error reporting. We should surface them with
        # our exceptions.

        if 'errors' in resp.json().keys():
            if isinstance(resp.json()['errors'], list):
                exception_message += '\n' + ','.join(resp.json()['errors'])
            else:
                for k, v in resp.json()['errors'].items():
                    exception_message += str('\n' + k + ': ' + ','.join(v))

        return exception_message

    def get_user(self, user_id):
        """
        Get a user.

        `Args:`
            user_id: int
                The user id of the record to get.
        `Returns`:
            User json object
        """

        return self._base_get(endpoint='user', entity_id=user_id,
                              exception_message='User not found')

    def get_user_fields(self):
        """
        Get list of valid user fields that can be passed with the
        :meth:`ActionKit.create_user` method.

        `Returns`:
            List of user fields
        """

        resp = self._base_get(endpoint='user/schema')

        return list(resp['fields'].keys())

    def create_user(self, email, **kwargs):
        """
        Create a user.

        `Args:`
            email: str
                Email for the user
            user_dict: dict
                Optional; Additional user fields
        `Returns:`
            User json object
        """

        return self._base_post(endpoint='user', exception_message='Could not create user',
                               email=email, **kwargs)

    def update_user(self, user_id, **kwargs):
        """
        Update a user.

        `Args:`
            user_id: int
                The user id of the person to update
            user_dict: dict
                A dictionary of fields to update for the user.
        `Returns:`
            ``None``
        """

        resp = self.conn.patch(self._base_endpoint('user', user_id), data=json.dumps(kwargs))
        logger.info(f'{resp.status_code}: {user_id}')

    def delete_user(self, user_id):
        """
        Delete a user.

        `Args:`
            user_id: int
                The user id of the person to delete
        `Returns:`
            ``None``
        """

        resp = self.conn.delete(self._base_endpoint('user', user_id))
        logger.info(f'{resp.status_code}: {user_id}')

    def get_campaign(self, campaign_id):
        """
        Get a campaign.

        `Args:`
            campaign_id: int
                The campaign id of the record.
        `Returns`:
            Campaign json object
        """

        return self._base_get(endpoint='campaign', entity_id=campaign_id,
                              exception_message='Campaign not found')

    def get_campaign_fields(self):
        """
        Get list of valid campaign fields that can be passed with the
        :meth:`ActionKit.create_campaign` and :meth:`ActionKit.update_campaign` methods.

        `Returns`:
            List of campaign fields
        """

        resp = self._base_get(endpoint='campaign/schema')
        return list(resp['fields'].keys())

    def create_campaign(self, name, **kwargs):
        """
        Create a campaign.

        `Args:`
            name: str
                The name of the campaign to create
        `Returns`:
            API location of new resource
        """

        return self._base_post(endpoint='campaign', exception_message='Could not create campaign',
                               name=name, **kwargs)

    def get_event_create_page(self, event_create_page_id):
        """
        Get a event create page.

        `Args:`
            event_create_page_id: int
                The event create page id of the record to get.
        `Returns`:
            Event create page json object
        """

        return self._base_get(endpoint='eventcreatepage', entity_id=event_create_page_id,
                              exception_message='Event create page not found')

    def get_event_create_page_fields(self):
        """
        Get list of event create page fields that can be passed with the
        :meth:`ActionKit.create_event_create_page`.

        `Returns`:
            List of event create page fields
        """

        resp = self._base_get(endpoint='eventcreatepage/schema')
        return list(resp['fields'].keys())

    def create_event_create_page(self, name, campaign_id, title, **kwargs):
        """
        Add an event page to a campaign.

        `Args:`
            campaign_id: int
                The campaign to assoicate page with
            name: str
                The name of the page to create
            title: str
                The title of the page to create
        `Returns`:
            API location of new resource
        """

        return self._base_post(endpoint='eventcreatepage',
                               exception_message='Could not create event create page',
                               campaign=f'/rest/v1/campaign/{campaign_id}/',
                               name=name,
                               title=title,
                               **kwargs)

    def get_event_create_form(self, event_create_form_id):
        """
        Get a event create form.

        `Args:`
            event_create_form_id: int
                The event create form id of the record to get.
        `Returns`:
            Event create form json object
        """

        return self._base_get(endpoint='eventcreateform', entity_id=event_create_form_id,
                              exception_message='Event create page not found')

    def get_event_create_form_fields(self):
        """
        Get list of valid event create form fields that can be passed with the
        :meth:`ActionKit.create_event_create_form` method.

        `Returns`:
            List of event create form fields
        """

        resp = self._base_get(endpoint='eventcreateform/schema')
        return list(resp['fields'].keys())

    def create_event_create_form(self, page_id, thank_you_text, **kwargs):
        """
        Create a event create form.

        `Args:`
            page_id: int
                The page to associate the form with
            thank_you_text: str
                Free form thank you text
        `Returns:`
            API location of new resource
        """

        return self._base_post(endpoint='eventcreateform',
                               exception_message='Could not event create form',
                               page=f'/rest/v1/eventcreatepage/{page_id}/',
                               thank_you_text=thank_you_text,
                               **kwargs)

    def get_event_signup_page(self, event_signup_page_id):
        """
        Get event signup page.

        `Args:`
            event_signup_page_id: int
                The event signup page id of the record to get.
        `Returns`:
            Event signup page json object
        """

        return self._base_get(endpoint='eventsignuppage', entity_id=event_signup_page_id,
                              exception_message='User page signup page not found')

    def get_event_signup_page_fields(self):
        """
        Get list of valid event signup page fields that can be passed with the
        :meth:`ActionKit.create_event_signup_page` method.

        `Returns`:
            List of event signup page fields
        """

        resp = self._base_get(endpoint='eventsignuppage/schema')
        return list(resp['fields'].keys())

    def create_event_signup_page(self, name, campaign_id, title, **kwargs):
        """
        Add an event signup page to a campaign.

        `Args:`
            campaign_id: int
                The campaign to assoicate page with
            name: str
                The name of the page to create
            title: str
                The title of the page to create
        `Returns`:
            API location of new resource
        """

        return self._base_post(endpoint='eventsignuppage',
                               exception_message='Could not create signup page',
                               campaign=f'/rest/v1/campaign/{campaign_id}/',
                               name=name,
                               title=title,
                               **kwargs)

    def get_event_signup_form(self, event_signup_form_id):
        """
        Get a user.

        `Args:`
            event_signup_form_id: str
                The event signup form id of the record to get.
        `Returns`:
            Event signup form json object
        """

        return self._base_get(endpoint='eventsignupform', entity_id=event_signup_form_id,
                              exception_message='User page signup form not found')

    def get_event_signup_form_fields(self):
        """
        Get list of valid event signup form fields that can be passed with the
        :meth:`ActionKit.create_event_signup_form` method.

        `Returns`:
            List of event signup form fields
        """

        resp = self._base_get(endpoint='eventsignupform/schema')
        return list(resp['fields'].keys())

    def create_event_signup_form(self, page_id, thank_you_text, **kwargs):
        """
        Create a event signup form.

        `Args:`
            page_id: int
                The page to associate the form with
            thank_you_text: str
                Free form thank you text
        `Returns:`
            API location of new resource
        """

        return self._base_post(endpoint='eventsignupform',
                               exception_message='Could not event create signup form',
                               page=f'/rest/v1/page/{page_id}/',
                               thank_you_text=thank_you_text,
                               **kwargs)

    def get_page_followup(self, page_followup_id):
        """
        Get a page followup.

        `Args:`
            page_followup_id: int
                The user id of the record to get.
        `Returns`:
            Page followup json object
        """

        return self._base_get(endpoint='pagefollowup', entity_id=page_followup_id,
                              exception_message='Page followup not found')

    def get_page_followup_fields(self):
        """
        Get list of valid page followup fields that can be passed with the
        :meth:`ActionKit.create_page_followup` method.

        `Returns`:
            List of page followup fields
        """

        resp = self._base_get(endpoint='pagefollowup/schema')
        return list(resp['fields'].keys())

    def create_page_followup(self, signup_page_id, url, **kwargs):
        """
        Add a page followup.

        `Args:`
            signup_page_id: int
                The signup page to associate the followup page with
            url: str
                URL of the folloup page
        `Returns`:
            API location of new resource
        """

        return self._base_post(endpoint='pagefollowup',
                               exception_message='Could not create page followup',
                               page=f'/rest/v1/eventsignuppage/{signup_page_id}/',
                               url=url,
                               **kwargs)

    def create_generic_action(self, page, email=None, ak_id=None, **kwargs):
        """
        Post a generic action. One of ``ak_id`` or ``email`` is a required argument.

        `Args:`
            page:
                The page to post the action. The page short name.
            email:
                The email address of the user to post the action.
            ak_id:
                The action kit id of the record.
            **kwargs:
                Optional arguments and fields that can sent. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/manual/api/rest/actionprocessing.html>`_.
        `Returns`:
            dict
                The response json
        """ # noqa: E501,E261

        if not email or ak_id:
            raise ValueError('One of email or ak_id is required.')

        return self._base_post(endpoint='action',
                               exception_message='Could not create action.',
                               email=email,
                               page=page,
                               return_full_json=True,
                               **kwargs)

    def bulk_upload_csv(self, csv_file, import_page,
                        autocreate_user_fields=False, user_fields_only=False):
        """
        Bulk upload a csv file of new users or user updates.
        If you are uploading a table object, use bulk_upload_table instead.
        See `ActionKit User Upload Documentation
             <https://roboticdogs.actionkit.com/docs/manual/api/rest/uploads.html>`
        Be careful that blank values in columns will overwrite existing data.

        If you get a 500 error, try sending a much smaller file (say, one row),
        which is more likely to return the proper 400 with a useful error message

        `Args:`
            import_page: str
                The page to post the action. The page short name.
            csv_file: str or buffer
                The csv (optionally zip'd) file path or a file buffer object
                A user_id or email column is required.
                ActionKit rejects files that are larger than 128M
            autocreate_user_fields: bool
              When True columns starting with "user_" will be uploaded as user fields.
              See the `autocreate_user_fields documentation
              <https://roboticdogs.actionkit.com/docs/manual/api/rest/uploads.html#create-a-multipart-post-request>`
            user_fields_only: bool
              When uploading only an email/user_id column and user_ user fields,
              ActionKit has a fast processing path.
              This doesn't work, if you upload a zipped csv though.
        `Returns`:
            dict
                success: whether upload was successful
                progress_url: an API URL to get progress on upload processing
                res: requests http response object
        """

        # self.conn defaults to JSON, but this has to be form/multi-part....
        upload_client = self._conn({'accepts': 'application/json'})
        if isinstance(csv_file, str):
            csv_file = open(csv_file, 'rb')

        res = upload_client.post(
            self._base_endpoint('upload'),
            files={'upload': csv_file},
            data={'page': import_page,
                  'autocreate_user_fields': int(autocreate_user_fields),
                  'user_fields_only': int(user_fields_only)})
        rv = {'res': res,
              'success': res.status_code == 201,
              'progress_url': res.headers.get('Location')}
        return rv

    def bulk_upload_table(self, table, import_page, autocreate_user_fields=0,
                          no_overwrite_on_empty=False):
        """
        Bulk upload a table of new users or user updates.
        See `ActionKit User Upload Documentation
             <https://roboticdogs.actionkit.com/docs/manual/api/rest/uploads.html>`
        Be careful that blank values in columns will overwrite existing data.
        Tables with only an identifying column (user_id/email) and user_ user fields
        will be fast-processed -- this is useful for setting/updating user fields.

        .. note::
            If you get a 500 error, try sending a much smaller file (say, one row),
            which is more likely to return the proper 400 with a useful error message

        `Args:`
            import_page: str
                The page to post the action. The page short name.
            table: Table Class
                A Table of user data to bulk upload
                A user_id or email column is required.
            autocreate_user_fields: bool
                When True columns starting with "user_" will be uploaded as user fields.
                See the `autocreate_user_fields documentation
                  <https://roboticdogs.actionkit.com/docs/manual/api/rest/uploads.html#create-a-multipart-post-request>`
            no_overwrite_on_empty: bool
                When uploading user data, ActionKit will, by default, take a blank value
                and overwrite existing data for that user.
                This can be undesirable, if the goal is to only send updates.
                Setting this to True will divide up the table into multiple upload
                batches, changing the columns uploaded based on permutations of
                empty columns.
        `Returns`:
            dict
                success: bool -- whether upload was successful (individual rows may not have been)
                results: [dict] -- This is a list of the full results.
                         progress_url and res for any results
        """
        import_page = check_env.check('ACTION_KIT_IMPORTPAGE', import_page)
        upload_tables = [table]
        if no_overwrite_on_empty:
            upload_tables = self._split_tables_no_empties(table)
        results = []
        for tbl in upload_tables:
            user_fields_only = int(not any([
                h for h in tbl.columns
                if h != 'email' and not h.startswith('user_')]))
            results.append(self.bulk_upload_csv(tbl.to_csv(),
                                                import_page,
                                                autocreate_user_fields=autocreate_user_fields,
                                                user_fields_only=user_fields_only))
        return {
            'success': all([r['success'] for r in results]),
            'results': results
        }

    def _split_tables_no_empties(self, table):
        table_groups = {}
        for row in table:
            blanks = tuple(k for k in table.columns
                           if row.get(k) in (None, ''))
            grp = table_groups.setdefault(blanks, [])
            grp.append(row)
        results = []
        for blanks, subset in table_groups.items():
            subset_table = Table(subset)
            if blanks:
                subset_table.table = subset_table.table.cutout(*blanks)
            results.append(subset_table)
        return results



from Newmode import Client
from parsons.utilities import check_env
from parsons.etl import Table
import logging

logger = logging.getLogger(__name__)


class Newmode:

    def __init__(self, api_user=None, api_password=None, api_version=None):
        """
        Args:
            api_user: str
                The Newmode api user. Not required if ``NEWMODE_API_USER`` env variable is
                passed.
            api_password: str
                The Newmode api password. Not required if ``NEWMODE_API_PASSWORD`` env variable is
                passed.
            api_version: str
                The Newmode api version. Defaults to "v1.0" or the value of ``NEWMODE_API_VERSION``
                env variable.
        Returns:
            Newmode class
        """
        self.api_user = check_env.check('NEWMODE_API_USER', api_user)
        self.api_password = check_env.check('NEWMODE_API_PASSWORD', api_password)

        if api_version is None:
            api_version = "v1.0"

        self.api_version = check_env.check('NEWMODE_API_VERSION', api_version)

        self.client = Client(api_user, api_password, api_version)

    def convert_to_table(self, data):
        # Internal method to create a Parsons table from a data element.
        table = None
        if type(data) is list:
            table = Table(data)
        else:
            table = Table([data])

        return table

    def get_tools(self, params={}):
        """
        Get existing tools.
        Args:
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Tools information as table.
        """
        tools = self.client.getTools(params=params)
        if tools:
            return self.convert_to_table(tools)
        else:
            logging.warning("Empty tools returned")
            return self.convert_to_table([])

    def get_tool(self, tool_id, params={}):
        """
        Get specific tool.
        Args:
            tool_id:
                The id of the tool to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Tool information.
        """
        tool = self.client.getTool(tool_id, params=params)
        if tool:
            return tool
        else:
            logging.warning("Empty tool returned")
            return None

    def lookup_targets(self, tool_id, search=None, params={}):
        """
        Lookup targets for a given tool
        Args:
            tool_id:
                The tool to lookup targets.
            search:
                The search criteria. It could be:
                - Empty: If empty, return custom targets associated to the tool.
                - Postal code: Return targets matched by postal code.
                - Lat/Long: Latitude and Longitude pair separated by '::'.
                Ex. 45.451596::-73.59912099999997. It will return targets
                matched for those coordinates.
                - Search term: For your csv tools, this will return targets
                matched by given valid search term.
        Returns:
            Targets information as table.
        """
        targets = self.client.lookupTargets(tool_id, search, params=params)
        if targets:
            data = []
            for key in targets:
                if key != '_links':
                    data.append(targets[key])
            return self.convert_to_table(data)
        else:
            logging.warning("Empty targets returned")
            return self.convert_to_table([])

    def get_action(self, tool_id, params={}):
        """
        Get the action information for a given tool.
        Args:
            tool_id:
                The id of the tool to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Tool action information.
        """
        action = self.client.getAction(tool_id, params=params)
        if action:
            return action
        else:
            logging.warning("Empty action returned")
            return None

    def run_action(self, tool_id, payload, params={}):
        """
        Run specific action with given payload.
        Args:
            tool_id:
                The id of the tool to run.
            payload:
                Payload data used to run the action. Structure will depend
                on the stuff returned by get_action.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Action link (if otl) or sid.
        """
        action = self.client.runAction(tool_id, payload, params=params)
        if action:
            if 'link' in action:
                return action['link']
            else:
                return action['sid']
        else:
            logging.warning("Error in response")
            return None

    def get_target(self, target_id, params={}):
        """
        Get specific target.
        Args:
            target_id:
                The id of the target to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Target information.
        """
        target = self.client.getTarget(target_id, params=params)
        if target:
            return target
        else:
            logging.warning("Empty target returned")
            return None

    def get_campaigns(self, params={}):
        """
        Get existing campaigns.
        Args:
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Campaigns information as table.
        """
        campaigns = self.client.getCampaigns(params=params)
        if campaigns:
            return self.convert_to_table(campaigns)
        else:
            logging.warning("Empty campaigns returned")
            return self.convert_to_table([])

    def get_campaign(self, campaign_id, params={}):
        """
        Get specific campaign.
        Args:
            campaign_id:
                The id of the campaign to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Campaign information.
        """
        campaign = self.client.getCampaign(campaign_id, params=params)
        if campaign:
            return campaign
        else:
            logging.warning("Empty campaign returned")
            return None

    def get_organizations(self, params={}):
        """
        Get existing organizations.
        Args:
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Organizations information as table.
        """
        organizations = self.client.getOrganizations(params=params)
        if organizations:
            return self.convert_to_table(organizations)
        else:
            logging.warning("Empty organizations returned")
            return self.convert_to_table([])

    def get_organization(self, organization_id, params={}):
        """
        Get specific organization.
        Args:
            organization_id:
                The id of the organization to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Organization information.
        """
        organization = self.client.getOrganization(organization_id, params=params)
        if organization:
            return organization
        else:
            logging.warning("Empty organization returned")
            return None

    def get_services(self, params={}):
        """
        Get existing services.
        Args:
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Services information as table.
        """
        services = self.client.getServices(params=params)
        if services:
            return self.convert_to_table(services)
        else:
            logging.warning("Empty services returned")
            return self.convert_to_table([])

    def get_service(self, service_id, params={}):
        """
        Get specific service.
        Args:
            service_id:
                The id of the service to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Service information.
        """
        service = self.client.getService(service_id, params=params)
        if service:
            return service
        else:
            logging.warning("Empty service returned")
            return None

    def get_outreaches(self, tool_id, params={}):
        """
        Get existing outreaches for a given tool.
        Args:
            tool_id:
                Tool to return outreaches.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Outreaches information as table.
        """
        outreaches = self.client.getOutreaches(tool_id, params=params)
        if outreaches:
            return self.convert_to_table(outreaches)
        else:
            logging.warning("Empty outreaches returned")
            return self.convert_to_table([])

    def get_outreach(self, outreach_id, params={}):
        """
        Get specific outreach.
        Args:
            outreach_id:
                The id of the outreach to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Outreach information.
        """
        outreach = self.client.getOutreach(outreach_id, params=params)
        if outreach:
            return outreach
        else:
            logging.warning("Empty outreach returned")
            return None



import unittest
import os
import requests_mock
from parsons.ngpvan import VAN
from test.utils import validate_list, assert_matching_tables

os.environ['VAN_API_KEY'] = 'SOME_KEY'

class TestActivistCodes(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters", raise_for_status=False)

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_activist_codes(self, m):

        # Create response
        json = {u'count': 43, u'items':
                [{u'status': u'Active',
                  u'scriptQuestion': None,
                  u'name': u'TEST CODE',
                  u'mediumName': u'TEST CODE',
                  u'activistCodeId': 4388538,
                  u'shortName': u'TC',
                  u'type': u'Action',
                  u'description': None}],
                u'nextPageLink': None}

        m.get(self.van.connection.uri + 'activistCodes', json=json)

        # Expected Structure
        expected = ['status', 'scriptQuestion', 'name', 'mediumName',
                    'activistCodeId', 'shortName', 'type', 'description']

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.van.get_activist_codes()))

        # To Do: Test what happens when it doesn't find any ACs

    @requests_mock.Mocker()
    def test_get_activist_code(self, m):

        # Create response
        json = {"status": "Active",
                "scriptQuestion": "null",
                "name": "Anti-Choice",
                "mediumName": "Anti",
                "activistCodeId": 4135099,
                "shortName": "AC",
                "type": "Constituency",
                "description": "A person who has been flagged as anti-choice."}

        m.get(self.van.connection.uri + 'activistCodes/4388538', json=json)

        self.assertEqual(json, self.van.get_activist_code(4388538))

    @requests_mock.Mocker()
    def test_toggle_activist_code(self, m):

        # Test apply activist code
        m.post(self.van.connection.uri + f'people/2335282/canvassResponses', status_code=204)
        self.assertTrue(self.van.toggle_activist_code(2335282, 4429154, 'apply'), 204)

        # Test remove activist code
        m.post(self.van.connection.uri + f'people/2335282/canvassResponses', status_code=204)
        self.assertTrue(self.van.toggle_activist_code(2335282, 4429154, 'remove'), 204)

    @requests_mock.Mocker()
    def test_apply_activist_code(self, m):

        # Test apply activist code
        m.post(self.van.connection.uri + f'people/2335282/canvassResponses', status_code=204)
        self.assertEqual(self.van.apply_activist_code(2335282, 4429154), 204)

    @requests_mock.Mocker()
    def test_remove_activist_code(self, m):

        # Test remove activist code
        m.post(self.van.connection.uri + f'people/2335282/canvassResponses', status_code=204)
        self.assertEqual(self.van.remove_activist_code(2335282, 4429154), 204)



from parsons.targetsmart.targetsmart_api import TargetSmartAPI
from parsons.targetsmart.targetsmart_automation import TargetSmartAutomation

__all__ = [
    'TargetSmartAPI',
    'TargetSmartAutomation'
]



import json
import logging
import requests
import time

from parsons.etl.table import Table
from parsons.utilities.check_env import check

logger = logging.getLogger(__name__)


class RedashTimeout(Exception):
    pass


class RedashQueryFailed(Exception):
    pass


class Redash(object):
    """
    Instantiate Redash Class

    `Args:`
        base_url: str
            The base url for your redash instance (excluding the final /)
        user_api_key: str
            The user API key found in the User's profile screen
        pause_time int
            Specify time between polling for refreshed queries (Defaults to 3 seconds)
        verify: bool
            For https requests, should the certificate be verified (Defaults to True)
    `Returns:`
        Redash Class
    """

    def __init__(self,
                 base_url=None,
                 user_api_key=None,
                 pause_time=3,
                 timeout=0,  # never timeout
                 verify=True):
        self.base_url = check('REDASH_BASE_URL', base_url)
        self.user_api_key = check('REDASH_USER_API_KEY', user_api_key, optional=True)
        self.pause = int(check('REDASH_PAUSE_TIME', pause_time, optional=True))
        self.timeout = int(check('REDASH_TIMEOUT', timeout, optional=True))

        self.verify = verify  # for https requests
        self.session = requests.Session()
        if user_api_key:
            self.session.headers.update({'Authorization': f'Key {user_api_key}'})

    def _poll_job(self, session, job, query_id):
        start_secs = time.time()
        while job['status'] not in (3, 4):
            if self.timeout and start_secs + self.timeout < time.time():
                raise RedashTimeout(f'Redash timeout: {self.timeout}')
            poll_url = '{}/api/jobs/{}'.format(self.base_url, job['id'])
            response = session.get(poll_url, verify=self.verify)
            response_json = response.json()
            job = response_json.get(
                'job',
                {'status': 'Error NO JOB IN RESPONSE: {}'.format(json.dumps(response_json))})
            logger.debug("poll url:%s id:%s status:%s err:%s",
                         poll_url, query_id, job['status'], job.get('error'))
            time.sleep(self.pause)

        if job['status'] == 3:  # 3 = completed
            return job['query_result_id']
        elif job['status'] == 4:  # 3 = ERROR
            raise RedashQueryFailed('Redash Query {} failed: {}'.format(query_id, job['error']))

    def get_fresh_query_results(self, query_id=None, params=None):
        """
        Make a fresh query result and get back the CSV http response object back
        with the CSV string in result.content

        `Args:`
            query_id: str or int
                The query id of the query
            params: dict
                If there are values for the redash query parameters
                (described https://redash.io/help/user-guide/querying/query-parameters
                e.g. "{{datelimit}}" in the query),
                then this is a dict that will pass the parameters in the POST.
                We add the "p_" prefix for parameters, so if your query had ?p_datelimit=....
                in the url, you should just set 'datelimit' in params here.
                If you set this with REDASH_QUERY_PARAMS environment variable instead of passing
                the values, then you must include the "p_" prefixes and it should be a single
                url-encoded string as you would see it in the URL bar.
        `Returns:`
            Table Class
        """
        query_id = check('REDASH_QUERY_ID', query_id, optional=True)
        params_from_env = check('REDASH_QUERY_PARAMS', '', optional=True)
        redash_params = ({'p_%s' % k: str(v).replace("'", "''") for k, v in params.items()}
                         if params else {})

        response = self.session.post(
            f'{self.base_url}/api/queries/{query_id}/refresh?{params_from_env}',
            params=redash_params,
            verify=self.verify)

        if response.status_code != 200:
            raise RedashQueryFailed(f'Refresh failed for query {query_id}. {response.text}')

        job = response.json()['job']
        result_id = self._poll_job(self.session, job, query_id)
        if result_id:
            response = self.session.get(
                f'{self.base_url}/api/queries/{query_id}/results/{result_id}.csv',
                verify=self.verify)
            if response.status_code != 200:
                raise RedashQueryFailed(
                    f'Failed getting results for query {query_id}. {response.text}')
        else:
            raise RedashQueryFailed(f'Failed getting result {query_id}. {response.text}')
        return Table.from_csv_string(response.text)

    def get_cached_query_results(self, query_id=None, query_api_key=None):
        """
        Get the results from a cached query result and get back the CSV http response object back
        with the CSV string in result.content

        `Args:`
            query_id: str or int
                The query id of the query
            query_api_key: str
                If you did not supply a user_api_key on the Redash object, then you can
                supply a query_api_key to get cached results back anonymously.
        `Returns:`
            Table Class
        """
        query_id = check('REDASH_QUERY_ID', query_id)
        query_api_key = check('REDASH_QUERY_API_KEY', query_api_key, optional=True)
        params = {}
        if not self.user_api_key and query_api_key:
            params['api_key'] = query_api_key
        response = self.session.get(f'{self.base_url}/api/queries/{query_id}/results.csv',
                                    params=params,
                                    verify=self.verify)
        if response.status_code != 200:
            raise RedashQueryFailed(f'Failed getting results for query {query_id}. {response.text}')
        return Table.from_csv_string(response.text)

    @classmethod
    def load_to_table(cls, refresh=True, **kwargs):
        """
        Fast classmethod so you can get the data all at once:
        tabledata = Redash.load_to_table(base_url='https://example.com', user_api_key='abc123',
                                         query_id=1001, params={'datelimit': '2020-01-01'})
        This instantiates the class and makes the appropriate query type (refresh or cached)
        based on which arguments are supplied.

        `Args:`
            base_url: str
                The base url for your redash instance (excluding the final /)
            query_id: str or int
                The query id of the query
            user_api_key: str
                The user API key found in the User's profile screen required for refresh queries
            query_api_key: str
                If you did not supply a user_api_key on the Redash object, then you can
                supply a query_api_key to get cached results back anonymously.
            pause_time int
                Specify time between polling for refreshed queries (Defaults to 3 seconds)
            verify: bool
                For https requests, should the certificate be verified (Defaults to True)
            refresh: bool
                Refresh results or cached. (Defaults to True unless a query_api_key IS supplied)
            params: dict
                For refresh queries, if there are parameters in the query,
                then this is a dict that will pass the parameters in the POST.
                We add the "p_" prefix for parameters, so if your query had ?p_datelimit=....
                in the url, you should just set 'datelimit' in params here.

        `Returns:`
            Table Class
        """
        initargs = {a: kwargs.get(a)
                    for a in ('base_url', 'user_api_key', 'pause_time', 'timeout', 'verify')
                    if a in kwargs}
        obj = cls(**initargs)
        if not refresh or kwargs.get('query_api_key'):
            return obj.get_cached_query_results(kwargs.get('query_id'), kwargs.get('query_api_key'))
        else:
            return obj.get_fresh_query_results(kwargs.get('query_id'), kwargs.get('params'))



from parsons.etl import Table
from requests import request
from parsons.utilities import check_env, json_format
import datetime
from parsons.hustle.column_map import LEAD_COLUMN_MAP
import logging

logger = logging.getLogger(__name__)

HUSTLE_URI = 'https://api.hustle.com/v1/'
PAGE_LIMIT = 1000


class Hustle(object):
    """
    Instantiate Hustle Class

    `Args:`
        client_id:
            The client id provided by Hustle. Not required if ``HUSTLE_CLIENT_ID`` env variable
            set.
        client_secret:
            The client secret provided by Hustle. Not required if ``HUSTLE_CLIENT_SECRET`` env
            variable set.
    `Returns:`
        Hustle Class
    """

    def __init__(self, client_id, client_secret):

        self.uri = HUSTLE_URI
        self.client_id = check_env.check('HUSTLE_CLIENT_ID', client_id)
        self.client_secret = check_env.check('HUSTLE_CLIENT_SECRET', client_secret)
        self.token_expiration = None
        self._get_auth_token(client_id, client_secret)

    def _get_auth_token(self, client_id, client_secret):
        # Generate a temporary authorization token

        data = {'client_id': client_id,
                'client_secret': client_secret,
                'grant_type': 'client_credentials'}

        r = request('POST', self.uri + 'oauth/token', data=data)
        logger.debug(r.json())

        self.auth_token = r.json()['access_token']
        self.token_expiration = datetime.datetime.now() + datetime.timedelta(seconds=7200)
        logger.info("Authentication token generated")

    def _token_check(self):
        # Tokens are only valid for 7200 seconds. This checks to make sure that it has
        # not expired and generate another one if it has.

        logger.debug("Checking token expiration.")
        if datetime.datetime.now() >= self.token_expiration:

            logger.info("Refreshing authentication token.")
            self._get_auth_token(self.client_id, self.client_secret)

        else:

            pass

    def _request(self, endpoint, req_type='GET', args=None, payload=None, raise_on_error=True):

        url = self.uri + endpoint
        self._token_check()

        headers = {'Authorization': f'Bearer {self.auth_token}'}

        parameters = {'limit': PAGE_LIMIT}

        if args:
            parameters.update(args)

        r = request(req_type, url, params=parameters, json=payload, headers=headers)

        self._error_check(r, raise_on_error)

        # If a single item return the dict
        if 'items' not in r.json().keys():

            return r.json()

        else:
            result = r.json()['items']

        # Pagination
        while r.json()['pagination']['hasNextPage'] == 'true':

            parameters['cursor'] = r.json['pagination']['cursor']
            r = request(req_type, url, params=parameters, headers=headers)
            self._error_check(r, raise_on_error)
            result.append(r.json()['items'])

        return result

    def _error_check(self, r, raise_on_error):
        # Check for errors

        if r.status_code in (200, 201):

            logger.debug(r.json())
            return None

        if raise_on_error:

            logger.info(r.json())
            r.raise_for_status()
            return None

        else:

            logger.info(r.json())
            return None

    def get_agents(self, group_id):
        """
        Get a list of agents.

        `Args:`
            group_id: str
                The group id.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self._request(f'groups/{group_id}/agents'))
        logger.info(f'Got {tbl.num_rows} agents from {group_id} group.')
        return tbl

    def get_agent(self, agent_id):
        """
        Get a single agent.

        `Args:`
            agent_id: str
                The agent id.
        `Returns:`
            dict
        """

        r = self._request(f'agents/{agent_id}')
        logger.info(f'Got {agent_id} agent.')
        return r

    def create_agent(self, group_id, name, full_name, phone_number, send_invite=False, email=None):
        """
        Create an agent.

        `Args:`
            group_id: str
                The group id to assign the agent.
            name: str
                The name of the agent.
            full_name: str
                The full name of the agent.
            phone_number: str
                The valid phone number of the agent.
            send_invite: boolean
                Send an invitation to the agent.
            email:
                The email address of the agent.
        `Returns:`
            dict
        """

        agent = {'name': name,
                 'fullName': full_name,
                 'phoneNumber': phone_number,
                 'sendInvite': send_invite,
                 'email': email}

        # Remove empty args in dictionary
        agent = json_format.remove_empty_keys(agent)

        logger.info(f'Generating {full_name} agent.')
        return self._request(f'groups/{group_id}/agents', req_type="POST", payload=agent)

    def update_agent(self, agent_id, name=None, full_name=None, send_invite=False):
        """
        Update an agent.

        `Args:`
            agent_id: str
                The agent id.
            name: str
                The name of the agent.
            full_name: str
                The full name of the agent.
            phone_number: str
                The valid phone number of the agent.
            send_invite: boolean
                Send an invitation to the agent.
        `Returns:`
            dict
        """

        agent = {'name': name,
                 'fullName': full_name,
                 'sendInvite': send_invite}

        # Remove empty args in dictionary
        agent = json_format.remove_empty_keys(agent)

        logger.info(f'Updating agent {agent_id}.')
        return self._request(f'agents/{agent_id}', req_type="PUT", payload=agent)

    def get_organizations(self):
        """
        Get organizations.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self._request('organizations'))
        logger.info(f'Got {tbl.num_rows} organizations.')
        return tbl

    def get_organization(self, organization_id):
        """
        Get a single organization.

        `Args:`
            organization_id: str
                The organization id.
        `Returns:`
            dict
        """

        r = self._request(f'organizations/{organization_id}')
        logger.info(f'Got {organization_id} organization.')
        return r

    def get_groups(self, organization_id):
        """
        Get a list of groups.

        `Args:`
            organization_id: str
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self._request(f'organizations/{organization_id}/groups'))
        logger.info(f'Got {tbl.num_rows} groups.')
        return tbl

    def get_group(self, group_id):
        """
        Get a single group.

        `Args:`
            group_id: str
                The group id.
        """

        r = self._request(f'groups/{group_id}')
        logger.info(f'Got {group_id} group.')
        return r

    def get_lead(self, lead_id):
        """
        Get a single lead..

        `Args`:
            lead_id: str
                The lead id.
        `Returns:`
            dict
        """

        r = self._request(f'leads/{lead_id}')
        logger.info(f'Got {lead_id} lead.')
        return r

    def get_leads(self, organization_id=None, group_id=None):
        """
        Get leads metadata. One of ``organization_id`` and ``group_id`` must be passed
        as an argument. If both are passed, an error will be raised.

        `Args:`
            organization_id: str
                The organization id.
            group_id: str
                The group id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if organization_id is None and group_id is None:
            raise ValueError('Either organization_id or group_id required.')

        if organization_id is not None and group_id is not None:
            raise ValueError('Only one of organization_id and group_id may be populated.')

        if organization_id:
            endpoint = f'organizations/{organization_id}/leads'
            logger.info(f'Retrieving {organization_id} organization leads.')
        if group_id:
            endpoint = f'groups/{group_id}/leads'
            logger.info(f'Retrieving {group_id} group leads.')

        tbl = Table(self._request(endpoint))
        logger.info(f'Got {tbl.num_rows} leads.')
        return tbl

    def create_lead(self, group_id, phone_number, first_name, last_name=None, email=None,
                    notes=None, follow_up=None, custom_fields=None, tag_ids=None):
        """

        Create a lead.

        `Args:`
            group_id: str
                The group id to assign the leads.
            first_name: str
                The first name of the lead.
            phone_number: str
                The phone number of the lead.
            last_name: str
                The last name of the lead.
            email: str
                The email address of the lead.
            notes: str
                The notes for the lead.
            follow_up: str
                Follow up for the lead.
            custom_fields: dict
                A dictionary of custom fields, with key as the value name, and
                value as the value.
            tag_ids: list
                A list of tag ids.
        `Returns:`
                ``None``
        """

        lead = {'firstName': first_name,
                'lastName': last_name,
                'email': email,
                'phoneNumber': phone_number,
                'notes': notes,
                'followUp': follow_up,
                'customFields': custom_fields,
                'tagIds': tag_ids
                }

        # Remove empty args in dictionary
        lead = json_format.remove_empty_keys(lead)
        logger.info(f'Generating lead for {first_name} {last_name}.')
        return self._request(f'groups/{group_id}/leads', req_type="POST", payload=lead)

    def create_leads(self, table, group_id=None):
        """
        Create multiple leads. All unrecognized fields will be passed as custom fields. Column
        names must map to the following names.

        .. list-table::
            :widths: 20 80
            :header-rows: 1

            * - Column Name
              - Valid Column Names
            * - first_name
              - ``first_name``, ``first``, ``fn``, ``firstname``
            * - last_name
              - ``last_name``, ``last``, ``ln``, ``lastname``
            * - phone_number
               - ``phone_number``, ``phone``, ``cell``, ``phonenumber``, ``cell_phone``
                ``cellphone``
            * - email
              - ``email``, ``email_address``, ``emailaddress``
            * - follow_up
              - ``follow_up``, ``followup``

        `Args:`
            table: Parsons table
                A Parsons table containing leads
            group_id:
                The group id to assign the leads. If ``None``, must be passed as a column
                value.
        `Returns:`
            A table of created ids with associated lead id.
        """

        table.map_columns(LEAD_COLUMN_MAP)

        arg_list = ['first_name', 'last_name', 'email', 'phone_number', 'follow_up',
                    'tag_id', 'group_id']

        created_leads = []

        for row in table:

            lead = {'group_id': group_id}
            custom_fields = {}

            # Check for column names that map to arguments, if not assign
            # to custom fields
            for k, v in row.items():
                if k in arg_list:
                    lead[k] = v
                else:
                    custom_fields[k] = v

            lead['custom_fields'] = custom_fields

            # Group Id check
            if not group_id and 'group_id' not in table.columns:
                raise ValueError('Group Id must be passed as an argument or a column value.')
            if group_id:
                lead['group_id'] == group_id

            created_leads.append(self.create_lead(**lead))

        logger.info(f"Created {table.num_rows} leads.")
        return Table(created_leads)

    def update_lead(self, lead_id, first_name=None, last_name=None, email=None,
                    global_opt_out=None, notes=None, follow_up=None, tag_ids=None):
        """
        Update a lead.

        `Args`:
            lead_id: str
                The lead id
            first_name: str
                The first name of the lead
            last_name: str
                The last name of the lead
            email: str
                The email address of the lead
            global_opt_out: boolean
                Opt out flag for the lead
            notes: str
                The notes for the lead
            follow_up: str
                Follow up for the lead
        `Returns:`
            dict
        """

        lead = {'leadId': lead_id,
                'firstName': first_name,
                'lastName': last_name,
                'email': email,
                'globalOptedOut': global_opt_out,
                'notes': notes,
                'followUp': follow_up,
                'tagIds': tag_ids}

        # Remove empty args in dictionary
        lead = json_format.remove_empty_keys(lead)

        logger.info('Updating lead for {first_name} {last_name}.')
        return self._request(f'leads/{lead_id}', req_type="PUT", payload=lead)

    def get_tags(self, organization_id):
        """
        Get an organization's tags.

        `Args:`
            organization_id: str
                The organization id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self._request(f'organizations/{organization_id}/tags'))
        logger.info(f'Got {tbl.num_rows} tags for {organization_id} organization.')
        return tbl

    def get_tag(self, tag_id):
        """
        Get a single tag.

        `Args:`
            tag_id: str
                The tag id.
        `Returns:`
            dict
        """

        r = self._request(f'tags/{tag_id}')
        logger.info(f'Got {tag_id} tag.')
        return r



import unittest
from test.utils import assert_matching_tables

from parsons.etl.table import Table
from parsons.aws.aws_async import import_and_get_task, get_func_task_path
from parsons.aws.lambda_distribute import distribute_task


def fake_func(arg1, arg2, kwarg1=None, kwarg2=56):
    return (arg1, arg2, kwarg1, kwarg2)


count = 0
tableargs = (None, None)


def fake_table_process(table, **fakekwargs):
    global count
    global tableargs
    count = count + 1
    tableargs = (table, fakekwargs)


class FakeRunner(object):

    def __init__(self, init1=None):
        self.init1 = init1

    def foobar(self, table, a, x=None, y=None):
        global count
        global tableargs
        count = count + 1
        tableargs = (table, self.init1, a, x, y)
        if a == 'raise':
            raise Exception('foo bar')


class TestAsync(unittest.TestCase):

    def test_task_path_conversion(self):
        # fake_func
        fake_str = get_func_task_path(fake_func)
        print('fake_str', fake_str)
        fake_renewed = import_and_get_task(fake_str)
        self.assertEqual(fake_renewed(1, 2, 3), (1, 2, 3, 56))

        # Table.from_csv_string @classmethod
        csv_str = get_func_task_path(Table.from_csv_string, Table)
        print('csv_str', csv_str)
        csv_renewed = import_and_get_task(csv_str)
        assert_matching_tables(csv_renewed('x,y\n1,2'),
                               Table([('x', 'y'), (1, 2)]))

        # Table.to_dicts (instance)
        dicts_str = get_func_task_path(Table.to_dicts, Table)
        print('dicts_str', dicts_str)
        dicts_renewed = import_and_get_task(dicts_str,
                                            {'lst': [('x', 'y'), (1, 2)]})
        self.assertEqual(dicts_renewed(),
                         [{'x': 1, 'y': 2}])

    def test_distribute_task(self):
        global count
        global tableargs
        datatable = [
            ('x', 'y'),
            (1, 2), (3, 4), (5, 6), (7, 8), (9, 0),
            (11, 12), (13, 14), (15, 16), (17, 18), (19, 10),
        ]
        count = 0
        tableargs = None
        distribute_task(Table(datatable),
                        fake_table_process,
                        'foo',  # bucket
                        group_count=5,
                        storage='local',
                        func_kwargs={'x': 1, 'y': [2, 3]})
        self.assertEqual(count, 2)
        assert_matching_tables(
            tableargs[0],
            Table([('x', 'y'),
                   (11, 12), (13, 14), (15, 16), (17, 18), (19, 10)]))
        count = 0
        tableargs = None
        distribute_task(Table(datatable + [(0, 0)]),
                        FakeRunner.foobar,
                        'foo',  # bucket
                        group_count=5,
                        storage='local',
                        func_class=FakeRunner,
                        func_class_kwargs={'init1': 'initx'},
                        func_kwargs={'a': 1, 'x': 2, 'y': 3})
        self.assertEqual(count, 3)
        self.assertEqual(tableargs[1:], ('initx', 1, 2, 3))
        self.assertEqual(tableargs[1:], ('initx', 1, 2, 3))
        assert_matching_tables(
            tableargs[0],
            Table([('x', 'y'), (0, 0)]))

        # 3. catch=True (with throwing)
        count = 0
        tableargs = None
        distribute_task(Table(datatable[:6]),
                        FakeRunner.foobar,
                        'foo',  # bucket
                        group_count=5,
                        storage='local',
                        func_class=FakeRunner,
                        func_class_kwargs={'init1': 'initx'},
                        catch=True,
                        func_kwargs={'a': 'raise', 'x': 2, 'y': 3})



from base64 import b64encode, b64decode
import click
import json
import os

PREFIX = 'PRSNSENV'


def decode_credential(credential, save_path=None, export=True, echo=False):
    """Decode an encoded credential to a Python object.

    `Args:`
        credential: str
            An encoded credential.
        save_path: str
            The path for where to save the decoded credential.
        export: bool
            A flag for whether to export the decoded object to the environment.
            Defaults to true.
        echo: bool
            A flag for whether to print the decoded object. Defaults to False.
    `Returns:`
        dict
            The decoded object.
    """
    x = len(PREFIX)
    if credential[:x] != PREFIX:
        raise ValueError("Invalid Parsons variable.")

    decoded_str = b64decode(
        bytes(credential.replace(PREFIX, ""), "utf-8")).decode("utf-8")

    decoded_dict = json.loads(decoded_str)

    if save_path:
        with open(save_path, 'w') as f:
            f.write(json.dumps(decoded_dict))

    if export:
        for key, val in decoded_dict.items():
            os.environ[key] = str(val)

    if echo:
        print(decoded_dict)

    return decoded_dict


def encode_from_json_str(credential):
    """Encode credential(s) from a json string.

    `Args:`
        credential: str
            The credential json string to be encoded.
    `Returns:`
        str
            The encoded credential.
    """
    data = json.loads(credential)

    json_str = json.dumps(data)
    encoded_str = PREFIX + b64encode(bytes(json_str, "utf-8")).decode("utf-8")

    return encoded_str


def encode_from_json_file(credential_file):
    """Encode credential(s) from a json file.

    `Args:`
        credential_file: str
            The path to the json file with the credential to be encoded.
    `Returns:`
        str
            The encoded credential.
    """
    with open(credential_file, 'r') as f:
        data = json.load(f)

    json_str = json.dumps(data)
    encoded_str = PREFIX + b64encode(bytes(json_str, "utf-8")).decode("utf-8")

    return encoded_str


def encode_from_env(env_variables):
    """Encode credential(s) from the current environment.

    `Args:`
        env_variables: list
            The list of credentials from the environment to be encoded.
    `Returns:`
        str
            The encoded credential.
    """
    data = {}
    for var in env_variables:
        data[var] = os.environ[var]

    json_str = json.dumps(data)
    encoded_str = PREFIX + b64encode(bytes(json_str, "utf-8")).decode("utf-8")

    return encoded_str


def encode_from_dict(credential):
    """Encode credential(s) from a dictionary.

    `Args:`
        credential: dict
            The list of credentials from the environment to be encoded.
    `Returns:`
        str
            The encoded credential.
    """
    data_str = json.dumps(credential)
    encoded_str = PREFIX + b64encode(bytes(data_str, "utf-8")).decode("utf-8")

    return encoded_str


@click.command(options_metavar='[-e [-f] | -d [-xp] [-o <file>]]')
@click.argument('credential', metavar='credential')
@click.option('--encode', '-e', 'fn', flag_value='encode',
              default=True, help="Endcode a credential.")
@click.option('--decode', '-d', 'fn', flag_value='decode',
              help='Decode an encoded credential.')
@click.option('-f', 'is_file', is_flag=True, help=("Treat <credential> as a "
              "path to a file. Only valid with --encode."))
@click.option('-o', 'save_path', default='', metavar='<file>',
              help="The path for where to save the decoded credential.")
@click.option('-x', 'no_export', is_flag=True, default=False,
              help=("Do not export the variable to the environment. Only "
                    "valid with --decode."))
@click.option('-s', 'suppress', is_flag=True, default=False, help=("Suppress "
              "the output."))
def main(credential, fn, is_file=False, save_path="", no_export=False,
         suppress=False):
    """A command line tool to encode and decode credentials.

    Use this tool when the credentials for a service are split into multiple
    parts, and you'd rather deal with them as one variable. For example, to
    connect to a Redshift database, you need 5 parts: username, password,
    database name, host name, and port. With this tool, you could encode those
    5 pieces into one string variable, which can be stored as an environment
    variable, in another credential storage system like Civis's, or whatever
    you need.

    When encoding (-e, --encode) credentials, the valid types are json string,
    path to a josn file, or a list. If passing a list, comma-separate with no
    space, the names of the environment variables to encode.

    Encoding examples:
    \b
    # Encoding a json string. Note: this assumes the output of json.dumps(str).
    `python env_tools.py -e '{"key": "val", "key2": "val2"}'`

    \b
    # Encoding a json file.
    `python env_tools.py -e -f /path/to/credentials.json`

    \b
    # Encoding a list currenct environment variables.
    `python env_tools.py -e env_var1,env_var2,env_ var3`
    """
    if fn == 'encode':
        if is_file:
            enc_cred = encode_from_json_file(credential)
        else:
            try:
                cred = json.loads(credential)
                enc_cred = encode_from_dict(cred)
            except json.decoder.JSONDecodeError:
                cred = credential.split(',')
                enc_cred = encode_from_env(cred)
        if not suppress:
            print(enc_cred)
    elif fn == 'decode':
        decode_credential(credential, save_path, not no_export, not suppress)
    else:
        raise ValueError("Invalid function selected. Use --help for help.")


if __name__ == '__main__':
    main()



from suds.client import Client
import logging
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

URI = 'https://api.securevan.com/v4/'
SOAP_URI = 'https://api.securevan.com/Services/V3/ListService.asmx?WSDL'


class VANConnector(object):

    def __init__(self, api_key=None, auth_name='default', db=None):

        self.api_key = check_env.check('VAN_API_KEY', api_key)

        if db == 'MyVoters':
            self.db_code = 0
        elif db in ['MyMembers', 'MyCampaign', 'EveryAction']:
            self.db_code = 1
        else:
            raise KeyError('Invalid database type specified. Pick one of:'
                           ' MyVoters, MyCampaign, MyMembers, EveryAction.')

        self.uri = URI
        self.db = db
        self.auth_name = auth_name
        self.pagination_key = 'nextPageLink'
        self.auth = (self.auth_name, self.api_key + '|' + str(self.db_code))
        self.api = APIConnector(self.uri, auth=self.auth, data_key='items',
                                pagination_key=self.pagination_key)

        # We will not create the SOAP client unless we need to as this triggers checking for
        # valid credentials. As not all API keys are provisioned for SOAP, this keeps it from
        # raising a permission exception when creating the class.
        self._soap_client = None

    @property
    def api_key_profile(self):
        """
        Returns the API key profile with includes permissions and other metadata.
        """

        return self.get_request('apiKeyProfiles')[0]

    @property
    def soap_client(self):

        if not self._soap_client:

            # Create the SOAP client
            soap_auth = {'Header': {'DatabaseMode': self.soap_client_db(), 'APIKey': self.api_key}}
            self._soap_client = Client(SOAP_URI, soapheaders=soap_auth)

        return self._soap_client

    def soap_client_db(self):
        """
        Parse the REST database name to the accepted SOAP format
        """

        if self.db == 'MyVoters':
            return 'MyVoterFile'
        if self.db == 'EveryAction':
            return 'MyCampaign'
        else:
            return self.db

    def get_request(self, endpoint, **kwargs):

        r = self.api.get_request(self.uri + endpoint, **kwargs)
        data = self.api.data_parse(r)

        # Paginate
        while isinstance(r, dict) and self.api.next_page_check_url(r):
            r = self.api.get_request(r[self.pagination_key], **kwargs)
            data.extend(self.api.data_parse(r))

        return data

    def post_request(self, endpoint, **kwargs):

        return self.api.post_request(self.uri + endpoint, **kwargs)

    def delete_request(self, endpoint, **kwargs):

        return self.api.delete_request(self.uri + endpoint, **kwargs)

    def patch_request(self, endpoint, **kwargs):

        return self.api.patch_request(self.uri + endpoint, **kwargs)

    def put_request(self, endpoint, **kwargs):

        return self.api.put_request(self.uri + endpoint, **kwargs)



import ast
import petl
import logging

logger = logging.getLogger(__name__)

# The following values are the minimum and maximum values for MySQL int
# types. https://dev.mysql.com/doc/refman/8.0/en/integer-types.html
SMALLINT_VALUES = [-32768, 32767]
MEDIUMINT_VALUES = [-8388608, 8388607]
INT_VALUES = [-2147483648, 2147483647]

# Additional padding to add on to the maximum column width to account
# for the addition of future data sets.
VARCHAR_PAD = .25


class MySQLCreateTable():

    def __init__(self):

        pass

    def data_type(self, val, current_type):
        # Evaluate the MySQL data type of a given value. Then return
        # the data type of the value, while also weighing what the
        # current type is.

        # Stop if the current type is already a varchar
        if current_type == 'varchar':
            return current_type

        # Convert to string to reevaluate data type. If the value is already
        # a string, is None, or is empty, with raise and error. These should
        # be considered varchars.
        try:
            t = ast.literal_eval(str(val))
        except ValueError:
            return 'varchar'
        except SyntaxError:
            return 'varchar'

        # If a float, then end
        if type(t) == float:
            return 'float'

        # If an int and current type is not a float, then determine type of int.
        if type(t) == int and current_type != 'float':
            # Make sure that it is a valid integer
            if not self.is_valid_integer(val):
                return 'varchar'
            # Use smallest possible int type above TINYINT
            if (SMALLINT_VALUES[0] < t < SMALLINT_VALUES[1]) and current_type not in ['int', 'bigint', 'mediumint']:  # noqa: E501
                return 'smallint'
            elif (MEDIUMINT_VALUES[0] < t < MEDIUMINT_VALUES[1]) and current_type not in ['int', 'bigint']:  # noqa: E501
                return 'mediumint'
            elif (INT_VALUES[0] < t < INT_VALUES[1]) and current_type not in ['bigint']:
                return 'int'
            else:
                return 'bigint'

        # Finally return the current type
        else:
            return current_type

    def is_valid_integer(self, val):

        # Valid ints in python can contain an underscore, but Redshift can't. This
        # checks to see if there is an underscore in the value and turns it into
        # a varchar if so.
        try:
            if '_' in val:
                return False

        except TypeError:
            return True

        # If it has a leading zero, we should treat it as a varchar, since it is
        # probably there for a good reason (e.g. zipcode)
        if val.isdigit():
            if val[0] == '0':
                return False

        return True

    def evaluate_column(self, column_rows):
        # Generate MySQL data types and widths for a column.

        col_width = 0
        col_type = None

        # Iterate through each row in the column
        for row in column_rows:

            # Get the MySQL data type
            col_type = self.data_type(row, col_type)

            # Calculate width if a varchar
            if col_type == 'varchar':
                row_width = len(str(row.encode('utf-8')))

                # Evaluate width vs. current max width
                if row_width > col_width:
                    col_width = row_width

        return col_type, int(col_width + (col_width * VARCHAR_PAD))

    def evaluate_table(self, tbl):
        # Generate a dict of MySQL column types and widths for all columns
        # in a table.

        table_map = []

        for col in tbl.columns:
            col_type, col_width = self.evaluate_column(tbl.column_data(col))
            col_map = {'name': col, 'type': col_type, 'width': col_width}
            table_map.append(col_map)

        return table_map

    def create_statement(self, tbl, table_name):
        # Generate create statement SQL for a given Parsons table.

        # Validate and rename column names if needed
        tbl.table = petl.setheader(tbl.table, self.columns_convert(tbl.columns))

        # Generate the table map
        table_map = self.evaluate_table(tbl)

        # Generate the column syntax
        column_syntax = []
        for c in table_map:
            if c['type'] == 'varchar':
                column_syntax.append(f"{c['name']} {c['type']}({c['width']}) \n")
            else:
                column_syntax.append(f"{c['name']} {c['type']} \n")

        # Generate full statement
        return f"CREATE TABLE {table_name} ( \n {','.join(column_syntax)});"

    def columns_convert(self, columns):
        # Update column values to conform with MySQL requirements for columns.

        updated_columns = []

        for idx, col in enumerate(columns):

            # Lowercase
            col = col.lower()

            # Strip whitespace
            col = col.strip()

            # Replace spaces with underscores
            col = col.replace(' ', '_')

            # If name is integer, prepend with "x_"
            if col.isdigit():
                col = f"x_{col}"

            # If column name is longer than 64 characters, truncate.
            if len(col) > 64:
                col = col[:63]

            # If column name is empty...
            if col == "":
                col = f"col_{idx}"

            updated_columns.append(col)

        return updated_columns



import csv
from io import TextIOWrapper, BytesIO, StringIO
import logging
import sys
import traceback
import time

from parsons.aws.aws_async import get_func_task_path, import_and_get_task, run as maybe_async_run
from parsons.aws.s3 import S3
from parsons.etl.table import Table
from parsons.utilities.check_env import check

logger = logging.getLogger(__name__)


class DistributeTaskException(Exception):
    pass


class TestStorage:

    def __init__(self):
        self.data = {}

    def put_object(self, bucket, key, object_bytes):
        self.data[key] = object_bytes

    def get_range(self, bucket, key, rangestart, rangeend):
        return self.data[key][rangestart:rangeend]


class S3Storage:
    """
    These methods are pretty specialized, so we keep them
    inside this file rather than s3.py
    """

    def __init__(self):
        self.s3 = S3()

    def put_object(self, bucket, key, object_bytes, **kwargs):
        return self.s3.client.put_object(Bucket=bucket, Key=key, Body=object_bytes, **kwargs)

    def get_range(self, bucket, key, rangestart, rangeend):
        """
        Gets an explicit byte-range of an S3 file
        """
        # bytes is INCLUSIVE for the rangeend parameter, unlike python
        # so e.g. while python returns 2 bytes for data[2:4]
        # Range: bytes=2-4 will return 3!! So we subtract 1
        response = self.s3.client.get_object(
            Bucket=bucket, Key=key,
            Range='bytes={}-{}'.format(rangestart, rangeend - 1))
        return response['Body'].read()


FAKE_STORAGE = TestStorage()
S3_TEMP_KEY_PREFIX = "Parsons_DistributeTask"


def distribute_task_csv(csv_bytes_utf8, func_to_run, bucket,
                        header=None,
                        func_kwargs=None,
                        func_class=None,
                        func_class_kwargs=None,
                        catch=False,
                        group_count=100,
                        storage='s3'):
    """
    The same as distribute_task, but instead of a table, the
    first argument is bytes of a csv encoded into utf8.
    This function is used by distribute_task() which you should use instead.
    """
    global FAKE_STORAGE
    func_name = get_func_task_path(func_to_run, func_class)
    row_chunks = csv_bytes_utf8.split(b'\n')
    cursor = 0
    row_ranges = []
    # gather start/end bytes for each row
    for rowc in row_chunks:
        rng = [cursor]
        cursor = cursor + len(rowc) + 1  # +1 is the \n character
        rng.append(cursor)
        row_ranges.append(rng)

    # group the rows and get start/end bytes for each group
    group_ranges = []
    # table csv writer appends a terminal \r\n, so we do len-1
    for grpstep in range(0, len(row_ranges) - 1, group_count):
        end = min(len(row_ranges) - 1, grpstep + group_count - 1)
        group_ranges.append((row_ranges[grpstep][0], row_ranges[end][1]))

    # upload data
    filename = hash(time.time())
    storagekey = f"{S3_TEMP_KEY_PREFIX}/{filename}.csv"
    groupcount = len(group_ranges)
    logger.debug(f'distribute_task_csv storagekey {storagekey} w/ {groupcount} groups')

    response = None
    if storage == 's3':
        response = S3Storage().put_object(bucket, storagekey, csv_bytes_utf8)
    else:
        response = FAKE_STORAGE.put_object(bucket, storagekey, csv_bytes_utf8)

    # start processes
    results = [
        maybe_async_run(
            process_task_portion,
            [bucket, storagekey, grp[0], grp[1], func_name, header,
             storage, func_kwargs, catch, func_class_kwargs],
            # if we are using local storage, then it must be run locally, as well
            # (good for testing/debugging)
            remote_aws_lambda_function_name='FORCE_LOCAL' if storage == 'local' else None
        )
        for grp in group_ranges]
    return {'DEBUG_ONLY': 'results may vary depending on context/platform',
            'results': results,
            'put_response': response}


def distribute_task(table, func_to_run,
                    bucket=None,
                    func_kwargs=None,
                    func_class=None,
                    func_class_kwargs=None,
                    catch=False,
                    group_count=100,
                    storage='s3'):
    """
    Distribute processing rows in a table across multiple AWS Lambda invocations.

    If you are running the processing of a table inside AWS Lambda, then you
    are limited by how many rows can be processed within the Lambda's time limit
    (at time-of-writing, maximum 15min).

    Based on experience and some napkin math, with
    the same data that would allow 1000 rows to be processed inside a single
    AWS Lambda instance, this method allows 10 MILLION rows to be processed.

    Rather than converting the table to SQS
    or other options, the fastest way is to upload the table to S3, and then
    invoke multiple Lambda sub-invocations, each of which can be sent a
    byte-range of the data in the S3 CSV file for which to process.

    Using this method requires some setup. You have three tasks:

    1. Define the function to process rows, the first argument, must take
       your table's data (though only a subset of rows will be passed)
       (e.g. `def task_for_distribution(table, **kwargs):`)
    2. Where you would have run `task_for_distribution(my_table, **kwargs)`
       instead call `distribute_task(my_table, task_for_distribution, func_kwargs=kwargs)
       (either setting env var S3_TEMP_BUCKET or passing a bucket= parameter)
    3. Setup your Lambda handler to include :py:meth:`parsons.aws.event_command`
       (or run and deploy your lambda with `Zappa <https://github.com/Miserlou/Zappa>`_)

    To test locally, include the argument `storage="local"` which will test
    the distribute_task function, but run the task sequentially and in local memory.

    A minimalistic example Lambda handler might look something like this:

    .. code-block:: python
       :emphasize-lines: 5,6

       from parsons.aws import event_command, distribute_task

       def process_table(table, foo, bar=None):
           for row in table:
               do_sloooooow_thing(row, foo, bar)

       def handler(event, context):
           ## ADD THESE TWO LINES TO TOP OF HANDLER:
           if event_command(event, context):
               return
           table = FakeDatasource.load_to_table(username='123', password='abc')
           # table is so big that running
           #   process_table(table, foo=789, bar='baz') would timeout
           # so instead we:
           distribute_task(table, process_table,
                           bucket='my-temp-s3-bucket',
                           func_kwargs={'foo': 789, 'bar': 'baz'})

    `Args:`
        table: Parsons Table
           Table of data you wish to distribute processing across Lambda invocations
           of `func_to_run` argument.
        func_to_run: function
           The function you want to run whose
           first argument will be a subset of table
        bucket: str
           The bucket name to use for s3 upload to process the whole table
           Not required if you set environment variable ``S3_TEMP_BUCKET``
        func_kwargs: dict
           If the function has other arguments to pass along with `table`
           then provide them as a dict here. They must all be JSON-able.
        func_class: class
           If the function is a classmethod or function on a class,
           then pass the pure class here.
           E.g. If you passed `ActionKit.bulk_upload_table`,
           then you would pass `ActionKit` here.
        func_class_kwargs: dict
           If it is a class function, and the class must be instantiated,
           then pass the kwargs to instantiate the class here.
           E.g. If you passed `ActionKit.bulk_upload_table` as the function,
           then you would pass {'domain': ..., 'username': ... etc} here.
           This must all be JSON-able data.
        catch: bool
           Lambda will retry running an event three times if there's an
           exception -- if you want to prevent this, set `catch=True`
           and then it will catch any errors and stop retries.
           The error will be in CloudWatch logs with string "Distribute Error"
           This might be important if row-actions are not idempotent and your
           own function might fail causing repeats.
        group_count: int
           Set this to how many rows to process with each Lambda invocation (Default: 100)
        storage: str
           Debugging option: Defaults to "s3". To test distribution locally without s3,
           set to "local".
    `Returns:`
        Debug information -- do not rely on the output, as it will change
        depending on how this method is invoked.
    """
    if storage not in ('s3', 'local'):
        raise DistributeTaskException(f'storage argument must be s3 or local')
    bucket = check('S3_TEMP_BUCKET', bucket)
    csvdata = StringIO()
    outcsv = csv.writer(csvdata)
    outcsv.writerows(table.table.data())
    return distribute_task_csv(csvdata.getvalue().encode('utf-8-sig'),
                               func_to_run,
                               bucket,
                               header=table.columns,
                               func_kwargs=func_kwargs,
                               func_class=func_class,
                               func_class_kwargs=func_class_kwargs,
                               catch=catch,
                               group_count=group_count,
                               storage=storage)


def process_task_portion(bucket, storagekey, rangestart, rangeend, func_name, header,
                         storage='s3', func_kwargs=None, catch=False,
                         func_class_kwargs=None):
    global FAKE_STORAGE

    logger.debug(f'process_task_portion func_name {func_name}, '
                 f'storagekey {storagekey}, byterange {rangestart}-{rangeend}')
    func = import_and_get_task(func_name, func_class_kwargs)
    if storage == 's3':
        filedata = S3Storage().get_range(bucket, storagekey, rangestart, rangeend)
    else:
        filedata = FAKE_STORAGE.get_range(bucket, storagekey, rangestart, rangeend)

    lines = list(csv.reader(TextIOWrapper(BytesIO(filedata), encoding='utf-8-sig')))
    table = Table([header] + lines)
    if catch:
        try:
            func(table, **func_kwargs)
        except Exception:
            # In Lambda you can search for '"Distribute Error"' in the logs
            type_, value_, traceback_ = sys.exc_info()
            err_traceback_str = '\n'.join(traceback.format_exception(type_, value_, traceback_))
            return {'Exception': 'Distribute Error',
                    'error': err_traceback_str,
                    'rangestart': rangestart,
                    'rangeend': rangeend,
                    'func_name': func_name,
                    'bucket': bucket,
                    'storagekey': storagekey}
    else:
        func(table, **func_kwargs)



from parsons.freshdesk.freshdesk import Freshdesk
import unittest
import requests_mock
from test.utils import assert_matching_tables
from test.test_freshdesk import expected_json

DOMAIN = 'myorg'
API_KEY = 'mykey'

class TestFreshdesk(unittest.TestCase):

    def setUp(self):

        self.fd = Freshdesk(DOMAIN, API_KEY)

    @requests_mock.Mocker()
    def test_get_agents(self, m):

    	# Test that agents are returned correctly.
        m.get(self.fd.uri + 'agents', json=expected_json.test_agent)
        tbl = self.fd.get_agents()

    @requests_mock.Mocker()
    def test_get_tickets(self, m):

        # Test that tickets are returned correctly.
        m.get(self.fd.uri + 'tickets', json=expected_json.test_ticket)
        tbl = self.fd.get_tickets() 

    @requests_mock.Mocker()
    def test_get_companies(self, m):

        # Test that tickets are returned correctly.
        m.get(self.fd.uri + 'companies', json=expected_json.test_company)
        tbl = self.fd.get_companies() 

    @requests_mock.Mocker()
    def test_get_contacts(self, m):

        # Test that tickets are returned correctly.
        m.get(self.fd.uri + 'contacts', json=expected_json.test_contact)
        tbl = self.fd.get_contacts() 



import civis
from parsons.etl.table import Table
from parsons.utilities import check_env


class CivisClient(object):
    """
    Instantiate the Civis class.

    `Args:`
        api_key: str
            The Civis api key.
        db: str or int
            The Civis Redshift database. Can be a database id or the name of the
            database.
        **kwargs: args
            Option settings for the client that are `described in the documentation <https://civis-python.readthedocs.io/en/stable/client.html#civis.APIClient>`_.
    `Returns:`
        Civis class
    """  # noqa: E501

    def __init__(self, db=None, api_key=None, **kwargs):

        self.api_key = check_env.check('CIVIS_API_KEY', api_key)
        self.db = check_env.check('CIVIS_DATABASE', db)
        self.client = civis.APIClient(api_key=api_key, **kwargs)
        """
        The Civis API client. Utilize this attribute to access to lower level and more
        advanced methods which might not be surfaced in Parsons. A list of the methods
        can be found by reading the Civis API client `documentation <https://civis-python.readthedocs.io/en/stable/client.html>`_.
        """  # noqa: E501

    def query(self, sql, preview_rows=10, polling_interval=None, hidden=True,
              wait=True):
        """
        Execute a SQL statement as a Civis query.

        Run a query that may return no results or where only a small
        preview is required. To execute a query that returns a large number
        of rows, see :func:`~civis.io.read_civis_sql`.

        `Args`
            sql: str
                The SQL statement to execute.
            preview_rows: int, optional
                The maximum number of rows to return. No more than 100 rows can be
                returned at once.
            polling_interval: int or float, optional
                Number of seconds to wait between checks for query completion.
            hidden: bool, optional
                If ``True`` (the default), this job will not appear in the Civis UI.
            wait: boolean
                If ``True``, will wait for query to finish executing before exiting
                the method.
        `Returns`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        fut = civis.io.query_civis(sql, self.db, preview_rows=preview_rows, polling_interval=None)

        if not wait:

            return None

        result = fut.result()

        if result['result_rows'] is None:

            return None

        result['result_rows'].insert(0, result['result_columns'])

        return Table(result['result_rows'])

    def table_import(self, table_obj, table, max_errors=None,
                     existing_table_rows='fail', diststyle=None, distkey=None,
                     sortkey1=None, sortkey2=None, wait=True, **civisargs):
        """
        Write the table to a Civis Redshift cluster. Additional key word
        arguments can passed to `civis.io.dataframe_to_civis()  <https://civis-python.readthedocs.io/en/v1.9.0/generated/civis.io.dataframe_to_civis.html#civis.io.dataframe_to_civis>`_ # noqa: E501

        `Args`
            table_obj: obj
                A Parsons Table object
            table: str
                The schema and table you want to upload to. E.g., 'scratch.table'. Schemas
                or tablenames with periods must be double quoted, e.g. 'scratch."my.table"'.
            api_key: str
                Your Civis API key. If not given, the CIVIS_API_KEY environment variable will be
                used.
            max_errors: int
                The maximum number of rows with errors to remove from the import before failing.
            existing_table_rows: str
                The behaviour if a table with the requested name already exists. One of
                `'fail'`, `'truncate'`, `'append'` or `'drop'`. Defaults to `'fail'`.
            diststyle: str
                The distribution style for the table. One of `'even'`, `'all'` or `'key'`.
            distkey: str
                The column to use as the distkey for the table.
            sortkey1: str
                The column to use as the sortkey for the table.
            sortkey2: str
                The second column in a compound sortkey for the table.
            wait: boolean
                Wait for write job to complete before exiting method.
        `Returns`
            ``None``
        """ # noqa: E501,E261

        fut = civis.io.dataframe_to_civis(table_obj.to_dataframe(), database=self.db,
                                          table=table, max_errors=max_errors,
                                          existing_table_rows=existing_table_rows,
                                          diststyle=diststyle, distkey=distkey,
                                          sortkey1=sortkey1, sortkey2=sortkey2,
                                          headers=True, **civisargs)

        if wait:

            fut.result()



"""NGPVAN Signups Endpoints"""
from parsons.etl.table import Table
import logging

logger = logging.getLogger(__name__)


class Signups(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_signups_statuses(self, event_id=None, event_type_id=None):
        """
        Get a list of valid signup statuses for a given event type
        or event. You must pass one of ``event_id`` or ``event_type_id``
        but not both.

        `Args:`
            event_id: int
                A valid event id.
            event_type_id: int
                A valid event type id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if event_id is None and event_type_id is None:
            raise ValueError('One of event_id or event_type_id must be populated')

        if event_id is not None and event_type_id is not None:
            raise ValueError('Event Id and Event Type ID may not BOTH be populated')

        if event_id:
            params = {'eventId': event_id}
        if event_type_id:
            params = {'eventTypeId': event_type_id}

        tbl = Table(self.connection.get_request('signups/statuses', params=params))
        logger.info(f'Found {tbl.num_rows} signups.')
        return tbl

    def get_person_signups(self, vanid):
        """
        Get the signup history of a person.

        `Args:`
            vanid: int
                A valid vanid associated with a person.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('signups', params={'vanID': vanid}))
        logger.info(f'Found {tbl.num_rows} signups for {vanid}.')
        return self._unpack_signups(tbl)

    def get_event_signups(self, event_id):
        """
        Get the signup history of an event.

        `Args:`
            event_id: int
                A valid event_id associated with an event
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('signups', params={'eventId': event_id}))
        logger.info(f'Found {tbl.num_rows} signups for event {event_id}.')
        return self._unpack_signups(tbl)

    def get_signup(self, event_signup_id):
        """
        Get a single signup object.

        `Args:`
            event_signup_id: int
                A valid event_signup_id associated with a signup.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.connection.get_request(f'signups/{event_signup_id}')
        logger.info(f'Found sign up {event_signup_id}.')
        return r

    def create_signup(self, vanid, event_id, shift_id, role_id, status_id, location_id):
        """
        Create a new signup for an event.

        `Args:`
            vanid: int
                A valid vanid of the person to signup for the event.
            event_id: int
                A valid event_id to associate the person with the event
            shift_id:
                A shift_id, associated with the event to assign the person
            role_id:
                A role_id, associated with the event to assign the person
            status_id:
                A status_id of the person
            location_id:
                A location_id for the event
        `Returns:`
            Int
                The event signup id
        """

        signup = {'person': {'vanId': vanid},
                  'event': {'eventId': event_id},
                  'shift': {'eventShiftId': shift_id},
                  'role': {'roleId': role_id},
                  'status': {'statusId': status_id},
                  'location': {'locationId': location_id}
                  }

        r = self.connection.post_request('signups', json=signup)
        logger.info('Signup {r} created.')
        return r

    def update_signup(self, event_signup_id, shift_id=None, role_id=None, status_id=None,
                      location_id=None):
        """
        Update a signup object. All of the kwargs will update the values associated
        with them.

        `Args:`
            event_signup_id: int
                A valid event signup id
            shift_id: int
                The shift_id to update
            role_id: int
                The role_id to update
            status_id: int
                The status_id to update
            location_id: int
                The location_id to update
        `Returns:`
            ``None``
        """

        #  Get the signup object
        signup = self.connection.get_request(f'signups/{event_signup_id}')

        # Update the signup object
        if shift_id:
            signup['shift'] = {'eventShiftId': shift_id}
        if role_id:
            signup['role'] = {'roleId': role_id}
        if status_id:
            signup['status'] = {'statusId': status_id}
        if location_id:
            signup['location'] = {'locationId': location_id}

        return self.connection.put_request(f'signups/{event_signup_id}', json=signup)

    def delete_signup(self, event_signup_id):
        """
        Delete a signup object

        `Args:`
            event_signup_id: int
                A valid event signup id
        `Returns:`
            ``None``
        """

        r = self.connection.delete_request(f'signups/{event_signup_id}')
        logger.info(f'Signup {event_signup_id} deleted.')
        return r

    def _unpack_signups(self, table):

        # Unpack all of the nested jsons
        table.unpack_dict('person', prepend=False)
        table.unpack_dict('status')
        table.unpack_dict('event')
        table.unpack_dict('shift')
        table.unpack_dict('role')
        table.unpack_dict('location')

        return table



import os
from setuptools import find_packages
from distutils.core import setup

THIS_DIR = os.path.abspath(os.path.dirname(__file__))


def main():
    with open(os.path.join(THIS_DIR, 'requirements.txt')) as reqs:
        requirements = reqs.read().strip().split('\n')

    setup(
        name="parsons",
        version='0.13.1',
        author="The Movement Cooperative",
        author_email="info@movementcooperative.org",
        url='https://github.com/movementcoop/parsons',
        keywords=['PROGRESSIVE', 'API', 'ETL'],
        packages=find_packages(),
        install_requires=requirements,
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7'
        ]
    )


if __name__ == "__main__":
    main()
